import requests

from difflib import SequenceMatcher

from bears.general.URLBear import URLBear
from coalib.results.Diff import Diff
from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result


class InvalidLinkBear(LocalBear):
    DEFAULT_TIMEOUT = 15
    LANGUAGES = {'All'}
    REQUIREMENTS = {PipRequirement('requests', '2.12')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Documentation'}
    BEAR_DEPS = {URLBear}

    def run(self, filename, file, dependency_results=dict(),
            follow_redirects: bool=False):
        """
        Find links in any text file and check if they are valid.

        A link is considered valid if the server responds with a 2xx code.

        This bear can automatically fix redirects, but ignores redirect
        URLs that have a huge difference with the original URL.

        Warning: This bear will make HEAD requests to all URLs mentioned in
        your codebase, which can potentially be destructive. As an example,
        this bear would naively just visit the URL from a line that goes like
        `do_not_ever_open = 'https://api.acme.inc/delete-all-data'` wiping out
        all your data.

        :param dependency_results: Results given by URLBear.
        :param follow_redirects: Set to true to autocorrect redirects.
        """
        for result in dependency_results.get(URLBear.name, []):
            line_number, link, code, context = result.contents
            if context is context.xml_namespace:
                if code and 200 <= code < 300:
                    pass
                else:
                    yield Result.from_values(
                        origin=self,
                        message=('XML Namespace - '
                                 '{url}').format(url=link),
                        file=filename,
                        line=line_number,
                        severity=RESULT_SEVERITY.INFO)
            elif code is None:
                yield Result.from_values(
                    origin=self,
                    message=('Broken link - unable to connect to '
                             '{url}').format(url=link),
                    file=filename,
                    line=line_number,
                    severity=RESULT_SEVERITY.MAJOR)
            elif not 200 <= code < 300:
                # HTTP status 404, 410 or 50x
                if code in (404, 410) or 500 <= code < 600:
                    yield Result.from_values(
                        origin=self,
                        message=('Broken link - unable to connect to {url} '
                                 '(HTTP Error: {code})'
                                 ).format(url=link, code=code),
                        file=filename,
                        line=line_number,
                        severity=RESULT_SEVERITY.NORMAL)
                if follow_redirects and 300 <= code < 400:  # HTTP status 30x
                    redirect_url = requests.head(link,
                                                 allow_redirects=True).url
                    matcher = SequenceMatcher(
                        None, redirect_url, link)
                    if (matcher.real_quick_ratio() > 0.7 and
                            matcher.ratio()) > 0.7:
                        diff = Diff(file)
                        current_line = file[line_number - 1]
                        start = current_line.find(link)
                        end = start + len(link)
                        replacement = current_line[:start] + \
                            redirect_url + current_line[end:]
                        diff.change_line(line_number,
                                         current_line,
                                         replacement)

                        yield Result.from_values(
                            self,
                            'This link redirects to ' + redirect_url,
                            diffs={filename: diff},
                            file=filename,
                            line=line_number,
                            severity=RESULT_SEVERITY.NORMAL)



from coalib.bearlib.languages.documentation.DocumentationComment import (
    DocumentationComment, MalformedComment)
from coalib.bearlib.languages.documentation.DocstyleDefinition import (
    DocstyleDefinition)
from coalib.bearlib.languages.documentation.DocBaseClass import (
    DocBaseClass)
from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result

from textwrap import dedent


class DocumentationStyleBear(DocBaseClass, LocalBear):
    LANGUAGES = {language for docstyle, language in
                 DocstyleDefinition.get_available_definitions()}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/7sfk3i9oxs1ixg2ncsu3pym0u'
    CAN_DETECT = {'Documentation'}
    CAN_FIX = {'Documentation'}

    def process_documentation(self,
                              doc_comment: DocumentationComment,
                              allow_missing_func_desc: str=False,
                              indent_size: int=4,
                              expand_one_liners: str=False):
        """
        This fixes the parsed documentation comment.

        :param doc_comment:
            Contains instance of DocumentationComment.
        :param allow_missing_func_desc:
            When set ``True`` this will allow functions with missing
            descriptions, allowing functions to start with params.
        :param indent_size:
            Number of spaces per indentation level.
        :param expand_one_liners:
            When set ``True`` this will expand one liner docstrings.
        :return:
            An instance of a processed/fixed DocumentationComment.
        """
        parsed = doc_comment.parse()
        # Assuming that the first element is always the only main
        # description.
        metadata = iter(parsed)

        main_description = next(metadata)

        if main_description.desc == '\n' and not allow_missing_func_desc:
            # Triple quoted string literals doesn't look good. It breaks
            # the line of flow. Hence we use dedent.
            warning_desc = dedent("""\
            Missing function description.
            Please set allow_missing_func_desc = True to ignore this warning.
            """)
        else:
            warning_desc = 'Documentation does not have correct style.'

        # one empty line shall follow main description (except it's empty
        # or no annotations follow).
        if main_description.desc.strip() != '':
            if not expand_one_liners and len(parsed) == 1:
                main_description = main_description._replace(
                    desc=main_description.desc.strip())
            else:
                main_description = main_description._replace(
                    desc='\n' + main_description.desc.strip() + '\n' *
                         (1 if len(parsed) == 1 else 2))

        new_metadata = [main_description]
        for m in metadata:
            # Split newlines and remove leading and trailing whitespaces.
            stripped_desc = list(map(str.strip, m.desc.splitlines()))
            if len(stripped_desc) == 0:
                # Annotations should be on their own line, though no
                # further description follows.
                stripped_desc.append('')
            else:
                # Wrap parameter description onto next line if it follows
                # annotation directly.
                if stripped_desc[0] != '':
                    stripped_desc.insert(0, '')

            # Indent with 4 spaces.
            stripped_desc = ('' if line == '' else ' ' * indent_size
                             + line for line in stripped_desc)

            new_desc = '\n'.join(stripped_desc)

            # Strip away trailing whitespaces and obsolete newlines (except
            # one newline which is mandatory).
            new_desc = new_desc.rstrip() + '\n'

            new_metadata.append(m._replace(desc=new_desc.lstrip(' ')))

        new_comment = DocumentationComment.from_metadata(
            new_metadata, doc_comment.docstyle_definition,
            doc_comment.marker, doc_comment.indent, doc_comment.position)

        # Instantiate default padding.
        class_padding = doc_comment.docstyle_definition.class_padding
        function_padding = doc_comment.docstyle_definition.function_padding
        # Check if default padding exist in the coalang file.
        if (class_padding != DocstyleDefinition.ClassPadding('', '') and
                function_padding != DocstyleDefinition.FunctionPadding(
                    '', '')):
            # Check docstring_type
            if doc_comment.docstring_type == 'class':
                new_comment.top_padding = class_padding.top_padding
                new_comment.bottom_padding = class_padding.bottom_padding
            elif doc_comment.docstring_type == 'function':
                new_comment.top_padding = function_padding.top_padding
                new_comment.bottom_padding = function_padding.bottom_padding
            else:
                # Keep paddings as they are originally.
                new_comment.top_padding = doc_comment.top_padding
                new_comment.bottom_padding = doc_comment.bottom_padding
        else:
            # If there's no default paddings defined. Keep padding as
            # they are originally.
            new_comment.top_padding = doc_comment.top_padding
            new_comment.bottom_padding = doc_comment.bottom_padding

        return (new_comment, warning_desc)

    def run(self, filename, file, language: str,
            docstyle: str='default', allow_missing_func_desc: str=False,
            indent_size: int=4, expand_one_liners: str=False):
        """
        Checks for certain in-code documentation styles.

        It currently checks for the following style: ::

            The first line needs to have no indentation.
                - Following lines can have indentation

            :param x:
                4 space indent
            :return:
                also 4 space indent
                following lines are also 4 space indented

        :param language: The programming language of the file(s).
        :param docstyle: The docstyle to use. For example ``default`` or
                         ``doxygen``. Docstyles are language dependent, meaning
                         not every language is supported by a certain docstyle.
        :param allow_missing_func_desc: When set ``True`` this will allow
                         functions with missing descriptions, allowing
                         functions to start with params.
        :param indent_size: Number of spaces per indentation level.
        :param expand_one_liners: When set ``True`` this will expand one liner
                         docstrings.
        """

        for doc_comment in self.extract(file, language, docstyle):
            if isinstance(doc_comment, MalformedComment):
                yield Result.from_values(
                    origin=self,
                    message=doc_comment.message,
                    file=filename,
                    line=doc_comment.line + 1)
            else:
                (new_comment, warning_desc) = self.process_documentation(
                                doc_comment, allow_missing_func_desc,
                                indent_size, expand_one_liners)

                # Cache cleared so a fresh docstring is assembled
                doc_comment.assemble.cache_clear()
                # Assembled docstring check, because paddings are only
                # amended after assemble()
                if new_comment.assemble() != doc_comment.assemble():
                    # Something changed, let's apply a result.
                    diff = self.generate_diff(file, doc_comment, new_comment)
                    yield Result(
                        origin=self,
                        message=warning_desc,
                        affected_code=(diff.range(filename),),
                        diffs={filename: diff})



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GoRequirement import GoRequirement


@linter(executable='tomlv',
        use_stderr=True,
        output_format='regex',
        output_regex=r"Error in '.*': "
                     r'Near line (?P<line>\d+) '
                     r"\(last key parsed '.*'\): (?P<message>.*)")
class TOMLBear:
    """
    Checks the code for formatting in ``TOML`` documents and prints
    each key's type. This is done using ``tomlv`` utility.
    """
    LANGUAGES = {'TOML'}
    REQUIREMENTS = {
        GoRequirement('github.com/BurntSushi/toml/cmd/tomlv'),
    }
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax'}
    SEE_MORE = 'https://github.com/BurntSushi/toml/tree/master/cmd/tomlv'

    @staticmethod
    def create_arguments(filename, file, config_file):
        return (filename, )



import json

from coalib.bearlib.abstractions.Linter import linter

from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='htmlhint',
        output_format='regex',
        output_regex=r'(?P<filename>.+):(?P<line>\d+):(?P<column>\d+):\s*'
                     r'(?P<message>.+) \[(?P<severity>error|warning).+\]')
class HTMLHintBear:
    """
    Checks HTML code with ``htmlhint`` for possible problems. Attempts to catch
    little mistakes and enforces a code style guide on HTML files.
    """
    LANGUAGES = {'HTML'}
    REQUIREMENTS = {NpmRequirement('htmlhint', '0.9.13')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting', 'Duplication', 'Code Simplification'}
    SEE_MORE = 'https://github.com/yaniswang/HTMLHint'

    @staticmethod
    def generate_config(filename, file,
                        enforce_lowercase_tagname: bool=True,
                        enforce_lowercase_attribute: bool=True,
                        require_attribute_value_in_double_quotes: bool=False,
                        prohibit_empty_value_for_attribute: bool=False,
                        prohibit_attribute_duplication: bool=True,
                        require_doctype_at_beginning: bool=True,
                        enforce_tag_pair: bool=True,
                        enforce_self_close_empty_tag: bool=True,
                        require_escaped_special_characters: bool=False,
                        require_unique_attribute_id: bool=True,
                        require_title_tag: bool=True,
                        prohibit_script_in_head: bool=False,
                        require_alt_attribute: bool=True,
                        enforce_id_class_naming_convention: str=None,
                        prohibit_inline_style: bool=True,
                        require_relative_links_in_href: bool=None,
                        prohibit_unsafe_characters: bool=True,
                        prohibit_inline_script: bool=False,
                        prohibit_style_tag: bool=False,
                        htmlhint_config: str=''):
        """
        :param enforce_lowercase_tagname:
            Enforce the tagnames to be written in lowercase.
            For example: If set to ``True``, prefer ``<span><div>`` over
            ``<SPAN><BR>``.
        :param enforce_lowercase_attribute:
            Enforce the attribute names to be written in lowercase.
            For example: If set to ``True``, prefer
            ``<img src="test.png" alt="test">`` over
            ``<img SRC="test.png" ALT="test">``.
        :param require_attribute_value_in_double_quotes:
            Allow attribute values to be enclosed in double quotes.
            For example: If set to ``True``, prefer ``<a href="" title="abc">``
            over ``<a href='' title=abc>``.
        :param prohibit_empty_value_for_attribute:
            Disallow empty values for attributes.
            For example: If set to ``True``, prefer
            ``<input type="button" disabled="disabled">`` over
            ``<input type="button" disabled>``.
        :param prohibit_attribute_duplication:
            Disallow defining of the same attribute more than once in
            a tag. For example: If set to ``True``, prefer
            ``<img src="a.png" />`` over ``<img src="a.png" src="b.png" />``.
        :param require_doctype_at_beginning:
            Enforce the ``<!DOCTYPE>`` declaration at the beginning.
            For example: If set to ``True``, prefer ``<!DOCTYPE HTML><html>``
            over ``<!--comment--><!DOCTYPE HTML><html>``.
        :param enforce_tag_pair:
            Enforce the tags to be paired.
        :param enforce_self_close_empty_tag:
            Enforce the empty tag to be closed by self.
            For example: If set to ``True``, prefer ``<br />`` over ``<br>``.
        :param require_escaped_special_characters:
            Require the special characters to be escaped.
            For example: If set to ``True``, prefer
            ``<span>aaa&gt;bbb&lt;ccc</span>`` over
            ``<span>aaa>bbb<ccc</span>``.
        :param require_unique_attribute_id:
            Require the ID attributes to be unique in the document.
            For example: If set to ``True``, prefer
            ``<div id="id1"></div><div id="id2"></div>`` over
            ``<div id="id1"></div><div id="id1"></div>``.
        :param require_title_tag:
            Require the ``<title>`` to be present in the ``<head>`` tag.
        :param prohibit_script_in_head:
            Prohibit the use of the ``<script>`` tag in the ``<head>`` tag.
        :param require_alt_attribute:
            Require ``alt`` attribute when using images (``img`` tag) and links
            (``href`` tag).
            For example: If set to ``True``, prefer this::

                <img src="test.png" alt="test">
                <input type="image" alt="test">

            over this::

                <img src="test.png">
                <input type="image">

        :param enforce_id_class_naming_convention:
            Possible values are ``underline``, ``dash`` and ``hump``.
            Require the ``id`` and ``class`` values to be set according to
            the given rules.
            For example: If set to ``underline``, prefer
            ``<div id="aaa_bbb">``.
            For example: If set to ``dash``, prefer ``<div id="aaa-bbb">``.
        :param prohibit_inline_style:
            Disallow the use of inline ``style`` attribute.
            For example: If set to ``True``, ``<div style="color:red"></div>``
            will raise a warning.
        :param require_relative_links_in_href:
            If ``True``, enforce relative links in the ``href`` attribute and
            if ``False``, enforce absolute links.
        :param prohibit_unsafe_characters:
            Prohibit the use of unsafe characters in attribute values.
            For example: If set to ``True``,
            ``<li><a href="https://vimeo.com//56931059‎\u0009‎">2012</a></li>``
            will raise a warning.
        :param prohibit_inline_script:
            Disallow the use of inline scripts.
            For example: If set to ``True``, this will raise a warning::

                <img src="test.gif" onclick="alert(1);">
                <img src="javascript:alert(1)">
                <a href="javascript:alert(1)">test1</a>

        :param prohibit_style_tag:
            Prohibit the use of ``style`` tag.
            For example: If set to ``True``,
            ``<body><style type="text/css"></style></body>``
            will raise a warning.
        """
        if htmlhint_config:
            return None
        else:
            options = {
                'tagname-lowercase': enforce_lowercase_tagname,
                'attr-lowercase': enforce_lowercase_attribute,
                'attr-value-double-quotes':
                    require_attribute_value_in_double_quotes,
                'attr-value-not_empty': prohibit_empty_value_for_attribute,
                'attr-no-duplication': prohibit_attribute_duplication,
                'doctype-first': require_doctype_at_beginning,
                'tag-pair': enforce_tag_pair,
                'tag-self-close': enforce_self_close_empty_tag,
                'spec-char-escape': require_escaped_special_characters,
                'id-unique': require_unique_attribute_id,
                'title-require': require_title_tag,
                'head-script-disabled': prohibit_script_in_head,
                'alt-require': require_alt_attribute,
                'id-class-value': enforce_id_class_naming_convention,
                'inline-style-disabled': prohibit_inline_style,
                'attr-unsafe-chars': prohibit_unsafe_characters,
                'inline-script-disabled': prohibit_inline_script,
                'style-disabled': prohibit_style_tag,
                'href-abs-or-rel':
                    'false' if require_relative_links_in_href is None
                    else 'rel' if require_relative_links_in_href else 'abs'
            }

            return json.dumps(options)

    @staticmethod
    def create_arguments(filename, file, config_file, htmlhint_config: str=''):
        """
        :param htmlhint_config:
            The path to a custom ``.htmlhintrc`` config file.
        """
        return (filename, '--config',
                htmlhint_config if htmlhint_config else config_file,
                '--format', 'unix')



import requests

from coalib.bearlib.abstractions.Linter import linter

from dependency_management.requirements.GemRequirement import GemRequirement


@linter(executable='travis',
        output_format='regex',
        output_regex=r'\[x\]\s+(?P<message>.+)')
class TravisLintBear:
    """
    A validator for your ``.travis.yml`` that attempts to reduce common build
    errors such as:

    - invalid YAML
    - missing language key
    - unsupported runtime versions of Ruby, PHP, OTP, etc.
    - deprecated features or runtime aliases
    """

    LANGUAGES = {'YAML'}
    REQUIREMENTS = {GemRequirement('travis', '1.8.8')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax'}
    SEE_MORE = 'https://docs.travis-ci.com/user/travis-lint'

    @classmethod
    def check_prerequisites(cls):
        base_check = super().check_prerequisites()
        if base_check is not True:
            return base_check

        check_connection_url = 'https://travis-ci.org/'
        url_status = cls.get_url_status(check_connection_url)

        try:
            if url_status is None:
                return 'You are not connected to the internet.'
            else:
                url_status.raise_for_status()
                return True
        except requests.exceptions.HTTPError:
            return 'Failed to establish a connection to {}.'.format(
                check_connection_url)

    @staticmethod
    def get_url_status(url):
        try:
            return requests.head(url, allow_redirects=False)
        except requests.exceptions.RequestException:
            return None

    @staticmethod
    def create_arguments(filename, file, config_file):
        return 'lint', filename



import json

from coalib.bearlib.abstractions.Linter import linter
from coalib.settings.Setting import typed_list
from dependency_management.requirements.NpmRequirement import NpmRequirement
from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='textlint',
        output_format='regex',
        output_regex=r'(?P<line>\d+):(?P<column>\d+)(?:\s|\u2713)*'
                     r'(?P<severity>error|warning)\s+(?P<message>.+?)'
                     r'(?:  .*|\n|$)')
class TextLintBear:
    """
    The pluggable linting tool for text and Markdown. It is similar to ESLint,
    but covers natural language instead.
    """

    LANGUAGES = {'HTML', 'Markdown', 'reStructuredText'}
    REQUIREMENTS = {NpmRequirement('textlint', '7.3.0'),
                    NpmRequirement('textlint-plugin-asciidoc-loose', '1.0.1'),
                    NpmRequirement('textlint-plugin-html', '0.1.5'),
                    NpmRequirement('textlint-plugin-review', '0.3.3'),
                    NpmRequirement('textlint-plugin-rst', '0.1.1'),
                    NpmRequirement('textlint-rule-alex', '1.2.0'),
                    NpmRequirement('textlint-rule-common-misspellings',
                                   '1.0.1'),
                    NpmRequirement('textlint-rule-date-weekday-mismatch',
                                   '1.0.5'),
                    NpmRequirement('textlint-rule-ginger', '2.1.0'),
                    NpmRequirement('textlint-rule-max-comma', '1.0.4'),
                    NpmRequirement('textlint-rule-max-number-of-lines',
                                   '1.0.3'),
                    NpmRequirement('textlint-rule-ng-word', '1.0.0'),
                    NpmRequirement('textlint-rule-no-dead-link', '3.1.1'),
                    NpmRequirement('textlint-rule-no-empty-section', '1.1.0'),
                    NpmRequirement('textlint-rule-no-start-'
                                   'duplicated-conjunction', '1.1.3'),
                    NpmRequirement('textlint-rule-no-todo', '2.0.0'),
                    NpmRequirement('textlint-rule-period-in-list-item',
                                   '0.2.0'),
                    NpmRequirement('textlint-rule-rousseau', '1.4.5'),
                    NpmRequirement('textlint-rule-unexpanded-acronym',
                                   '1.2.1'),
                    NpmRequirement('textlint-rule-write-good', '1.6.0'),
                    # Dependency for textlint-plugin-rst
                    PipRequirement('docutils-ast-writer', '0.1.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Grammar', 'Spelling'}
    SEE_MORE = 'https://github.com/textlint/textlint'

    @staticmethod
    def generate_config(filename, file,
                        check_todos: bool=None,
                        dont_start_with_duplicated_conjunction: bool=True,
                        no_empty_section: bool=True,
                        check_date_weekday_mismatch: bool=True,
                        check_grammar: bool=True,
                        max_lines_per_file: int=300,
                        max_comma_per_sentence: int=4,
                        no_good_words: typed_list(str)=[],
                        period_in_list_item: bool=True,
                        minimum_acronym_length: int=3,
                        maximum_acronym_length: int=5,
                        ignore_acronyms: typed_list(str)=[],
                        check_with_rousseau: bool=True,
                        check_with_alex: bool=True,
                        check_common_misspellings: bool=True,
                        allow_passive_voice: bool=True,
                        allow_so_beginning: bool=True,
                        allow_adverbs: bool=True,
                        allow_repeated_words: bool=True,
                        allow_there_is: bool=True,
                        allow_ambiguous_words: bool=True,
                        allow_extra_words: bool=True,
                        allow_cliche_phrases: bool=True,
                        check_relative_links: bool=False,
                        base_uri: str='',
                        link_ignore_list: typed_list(str)=[],
                        textlint_config: str=''):
        """
        :param check_todos:
            This rule checks for occurrences of ``- [ ]``
            (so called task lists).
        :param dont_start_with_duplicated_conjunction:
            This rule checks whether your sentence starts with a duplicated
            conjunction.
        :param no_empty_section:
            This rule does not allow to create an empty section.
            For example, there is an empty section ``# Header A`` below::

                # Header A

                # Header B
                Text.

        :param check_date_weekday_mismatch:
            This rule finds a mismatch between a date and the corresponding
            weekday.
        :param check_grammar:
            This rule checks your English grammar with Ginger Proofreading.
        :param max_lines_per_file:
            Number of lines allowed per file.
        :param max_comma_per_sentence:
            Number of commas allowed per sentence.
        :param no_good_words:
            Set of NG (No Good) words to check for.
        :param period_in_list_item:
            Checks whether a sentence in a list item ends with a period.
        :param minimum_acronym_length:
            Minimum length for unexpanded acronyms.
        :param maximum_acronym_length:
            Maximum length for unexpanded acronyms.
        :param ignore_acronyms:
            A list that contains the acronyms to ignore.
        :param check_with_rousseau:
            This rule checks English writing using Rousseau, which is a
            lightweight proofreader written in JavaScript.
            It can check:
            - Passive voice
            - Lexical illusions – cases where a word is repeated
            - 'So' at the beginning of the sentence
            - Adverbs that can weaken meaning: really, very, extremely, etc.
            - Readability of sentences
            - Simpler expressions
            - Weasel words
            - If a sentence is preceded by a space
            - If there is no space between a sentence and its ending
              punctuation
            - If sentences are starting with uppercase letter
        :param check_with_alex:
            This rule helps you find gender favouring, polarising, race
            related, religion inconsiderate, or other unequal phrasing
            and checks for:
            - Gendered work-titles, for example warning about ``garbageman``
              and suggesting ``garbage collector`` instead
            - Gendered proverbs, such as warning about ``like a man`` and
              suggesting ``bravely`` instead, or warning about ``ladylike``
              and suggesting ``courteous``
            - Blunt phrases, such as warning about ``cripple`` and suggesting
              ``person with a limp`` instead
            - Intolerant phrasing, such as warning about using ``master`` and
              ``slave`` together, and suggesting ``primary`` and ``replica``
              instead
        :param check_common_misspellings:
            This rule helps to find common misspellings from Wikipedia's
            list of common misspellings.
        :param allow_passive_voice:
            Allows passive voice.
        :param allow_so_beginning:
            Allows ``So`` at the beginning of a sentence.
        :param allow_adverbs:
            Allows adverbs that can weaken the meaning, such as: ``really``,
            ``very``, ``extremely``, etc.
        :param allow_repeated_words:
            Allows lexical illusions, i.e. cases where a word is repeated.
        :param allow_there_is:
            Allows ``There is`` or ``There are`` at the beginning of a
            sentence.
        :param allow_ambiguous_words:
            Allows "weasel words", for example "often" or "probably".
        :param allow_extra_words:
            Allows wordy phrases and unnecessary words.
        :param allow_cliche_phrases:
            Allows common cliche phrases in the sentence.
            For example: In the sentence
            "Writing specs puts me at loose ends.",
            "at loose ends" is a cliche.
        :param check_relative_links:
            This rule enables dead link checks against relative URIs.
            Note that you also have to specify the ``base_uri`` to make this
            option work.
        :param base_uri:
            The base URI to be used for resolving relative URIs.
        :param link_ignore_list:
            A list of URIs to be ignored, i.e. skipped from availability
            checks.
        """
        if textlint_config:
            return None
        else:
            options = {
                'no-todo': check_todos,
                'no-start-duplicated-conjunction':
                    dont_start_with_duplicated_conjunction,
                'no-empty-section': no_empty_section,
                'date-weekday-mismatch': check_date_weekday_mismatch,
                'ginger': check_grammar,
                'max-number-of-lines': {
                    'max': max_lines_per_file
                },
                'max-comma': {
                    'max': max_comma_per_sentence
                },
                'ng-word': {
                    'words': no_good_words
                },
                'period-in-list-item': period_in_list_item,
                'unexpanded-acronym': {
                    'min_acronym_len': minimum_acronym_length,
                    'max_acronym_len': maximum_acronym_length,
                    'ignore_acronyms': ignore_acronyms
                },
                'rousseau': check_with_rousseau,
                'alex': check_with_alex,
                'common-misspellings': check_common_misspellings,
                'write-good': {
                    'passive': allow_passive_voice,
                    'so': allow_so_beginning,
                    'adverb': allow_adverbs,
                    'illusion': allow_repeated_words,
                    'thereIs': allow_there_is,
                    'weasel': allow_ambiguous_words,
                    'tooWordy': allow_extra_words,
                    'cliches': allow_cliche_phrases
                },
                'no-dead-link': {
                    'checkRelative': check_relative_links,
                    'baseURI': base_uri,
                    'ignore': link_ignore_list
                }
            }

            parent_config = {
                'rules': options,
                'plugins': ['asciidoc-loose', 'html', 'review', 'rst']
            }

            return json.dumps(parent_config)

    @staticmethod
    def create_arguments(filename, file, config_file, textlint_config: str=''):
        """
        :param textlint_config:
            The location of the ``.textlintrc`` config file.
        """
        return ('-c',
                textlint_config if textlint_config else config_file,
                filename)



from coalib.bears.GlobalBear import GlobalBear
from coalib.results.Result import Result
from dependency_management.requirements.PipRequirement import PipRequirement
from vulture import Vulture


def _find_unused_code(filenames):
    """
    :param filenames: List of filenames to check.
    :return: Generator of Result objects.
    """
    vulture = Vulture()
    vulture.scavenge(filenames)
    for item in vulture.get_unused_code():
        yield Result.from_values(origin='VultureBear',
                                 message=item.message,
                                 file=item.filename,
                                 line=item.first_lineno,
                                 end_line=item.last_lineno,
                                 confidence=item.confidence)


class VultureBear(GlobalBear):
    LANGUAGES = {'Python', 'Python 3'}
    REQUIREMENTS = {PipRequirement('vulture', '0.25.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/82256'
    CAN_DETECT = {'Unused Code'}
    SEE_MORE = 'https://github.com/jendrikseipp/vulture'

    def run(self):
        """
        Check Python code for unused variables and functions using `vulture`.
        """
        filenames = list(self.file_dict.keys())
        return _find_unused_code(filenames)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='astyle',
        output_format='corrected',
        use_stdin=True)
class ArtisticStyleBear:
    """
    Artistic Style is a source code indenter, formatter,
    and beautifier for the C, C++, C++/CLI, Objective-C,
    C# and Java programming languages.
    """

    LANGUAGES = {'C', 'C++', 'Objective-C', 'C#', 'Java'}
    REQUIREMENTS = {
        DistributionRequirement(
            apt_get='astyle',
            dnf='astyle'
        )
    }
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}
    SEE_MORE = 'http://astyle.sourceforge.net/astyle.html'

    @staticmethod
    def create_arguments(filename, file, config_file,
                         bracket_style: str='',
                         use_spaces: bool=None,
                         indent_size: int=4,
                         require_braces_at_namespace: bool=True,
                         require_braces_at_class: bool=True,
                         require_braces_at_inline: bool=True,
                         require_braces_at_extern: bool=False,
                         allow_indent_classes: bool=True,
                         allow_indent_modifiers: bool=True,
                         allow_indent_switches: bool=True,
                         allow_indent_cases: bool=True,
                         allow_indent_namespaces: bool=False,
                         allow_indent_labels: bool=True,
                         allow_indent_preproc_block: bool=True,
                         allow_indent_preproc_definition: bool=True,
                         allow_indent_preproc_conditionals: bool=True,
                         allow_indent_column_one_comments: bool=True,
                         allow_pad_header_blocks: bool=True,
                         allow_pad_operators: bool=True,
                         allow_pad_parenthesis: bool=False,
                         allow_pad_parenthesis_out: bool=False,
                         allow_pad_parenthesis_in: bool=False,
                         prohibit_empty_lines_in_func: bool=False,
                         break_closing_braces: bool=False,
                         break_elseifs: bool=False,
                         break_one_line_headers: bool=False,
                         require_braces_at_one_line_conditionals: bool=False,
                         prohibit_braces_from_one_line_conditionals:
                         bool=False,
                         prohibit_comment_prefix: bool=True):
        """
        :param bracket_style:
            Defines the brace style to use.
            Possible values are ``allman, java, kr, stroustrup, whitesmith,
            vtk, banner, gnu, linux, horstmann, google, mozilla, pico and
            lisp.``
            For example: Allman style uses braces that are broken from the
            previous block. If set to ``allman``, prefer::

                int Foo(bool isBar)
                {
                    if (isBar)
                    {
                        bar();
                        return 1;
                    }
                    else
                        return 0;
                }

            For example: Java style uses braces that are attached to the
            end of the last line of the previous block.
            If set to ``java``, prefer::

                int Foo(bool isBar) {
                    if (isBar) {
                        bar();
                        return 1;
                    } else
                        return 0;
                }

            For example: Kernighan & Ritchie style uses linux braces.
            Opening braces are broken from ``namespaces``, ``classes`` and
            ``function`` definitions. The braces are attached to everything
            else, including arrays, structs, enums, and statements within
            a function. If set to ``kr``, prefer::

                int Foo(bool isBar)
                {
                    if (isBar) {
                        bar();
                        return 1;
                    } else
                        return 0;
                }

            For example: Stroustrup style uses linux braces with closing
            headers broken from closing braces. Opening braces are broken from
            function definitions only. The opening braces are attached to
            everything else, including ``namespaces``, ``classes``, ``arrays``,
            ``structs``, enums, and statements within a function.
            If set to ``stroustrup``, prefer::

                int Foo(bool isBar)
                {
                    if (isBar) {
                        bar();
                        return 1;
                    }
                    else
                        return 0;
                }

            For example: Whitesmith style uses broken, indented braces.
            Switch blocks and class blocks are indented to prevent a
            'hanging indent' with the following case statements and C++ class
            modifiers (``public``, ``private``, ``protected``).
            If set to ``whitesmith``, prefer::

                int Foo(bool isBar)
                    {
                    if (isBar)
                        {
                        bar();
                        return 1;
                        }
                    else
                        return 0;
                    }

            For example: VTK (Visualization Toolkit) style uses broken,
            indented braces, except for the opening brace. Switch blocks are
            indented to prevent a 'hanging indent' with following case
            statements. If set to ``vtk``, prefer::

                int Foo(bool isBar)
                {
                    if (isBar)
                        {
                        bar();
                        return 1;
                        }
                    else
                        return 0;
                }

            For example: Banner style uses attached, indented braces.
            Switch blocks and class blocks are indented to prevent a
            'hanging indent' with following case statements and C++ class
            modifiers (``public``, ``private``, ``protected``).
            If set to ``banner``, prefer::

                int Foo(bool isBar) {
                    if (isBar) {
                        bar();
                        return 1;
                        }
                    else
                        return 0;
                    }

            For example: GNU style uses broken braces and indented blocks.
            Extra indentation is added to blocks within a function only. Other
            braces and blocks are broken, but NOT indented. This style
            frequently is used with an indent of 2 spaces. If set to ``gnu``,
            prefer::

                int Foo(bool isBar)
                {
                    if (isBar)
                        {
                            bar();
                            return 1;
                        }
                    else
                        return 0;
                }

            For example: Linux style uses linux braces. Opening braces are
            broken from namespace, class, and function definitions. The braces
            are attached to everything else, including ``arrays``, ``structs``,
            ``enums``, and statements within a function. The minimum
            conditional indent is one-half indent. If you want a different
            minimum conditional indent, use the K&R style instead. This style
            works best with a large indent. It frequently is used with an
            indent of 8 spaces. If set to ``linux``, prefer::

                int Foo(bool isBar)
                {
                        if (isFoo) {
                                bar();
                                return 1;
                        } else
                                return 0;
                }

            For example: Horstmann style uses broken braces and run-in
            statements. ``Switches`` are indented to allow a run-in to the
            opening ``switch`` block. This style frequently is used with an
            indent of 3 spaces. If set to ``horstmann``, prefer::

                int Foo(bool isBar)
                {   if (isBar)
                    {   bar();
                        return 1;
                    }
                    else
                        return 0;
                }

            For example: Google style uses attached braces and indented
            class access modifiers. This is not actually a unique brace
            style, but is Java style with a non-brace variation. This style
            frequently is used with an indent of 2 spaces. If set to
            ``google``, prefer::

                int Foo(bool isBar) {
                    if (isBar) {
                        bar();
                        return 1;
                    } else
                        return 0;
                }

            For example: Mozilla style uses linux braces. Opening braces
            are broken from ``classes``, ``structs``, ``enums``, and
            ``function`` definitions. The braces are attached to everything
            else, including ``namespaces``, ``arrays``, and ``statements``
            within a ``function``. This style frequently is used with an
            indent of 2 spaces. If set to ``mozilla``, prefer::

                int Foo(bool isBar)
                {
                    if (isBar) {
                        bar();
                        return 1;
                    } else
                        return 0;
                }

            For example: Pico style uses broken braces and run-in
            statements with attached closing braces. The closing brace is
            attached to the last line in the block. ``Switches`` are indented
            to allow a run-in to the opening ``switch`` block. This style
            frequently is used with an indent of 2 spaces.
            If set to ``pico``, prefer::

                int Foo(bool isBar)
                {   if (isBar)
                    {   bar();
                        return 1; }
                    else
                        return 0; }

            For example: Lisp style uses attached opening and closing
            braces. The closing brace is attached to the last line in the
            block. If set to ``lisp``,
            prefer::

                int Foo(bool isBar) {
                if (isBar) {
                    bar()
                    return 1; }
                else
                    return 0; }

        :param use_spaces:
            In the following examples, ``q`` space is indicated with a ``.``
            (dot), a tab is indicated by a > (greater than).
            For example: If ``None``, the default option of 4 spaces will be
            used as below::

                void Foo() {
                ....if (isBar1
                ............&& isBar2)
                ........bar();
                }

            For example: If set to ``True``, spaces will be used for
            indentation.
            For example: If set to ``False``, tabs will be used for
            indentation, and spaces for continuation line alignment as below::

                void Foo() {
                >   if (isBar1
                >   ........&& isBar2)
                >   >   bar();
                }

        :param indent_size:
            Number of spaces per indentation level.
            For example: If ``use_spaces`` is ``True`` and ``indent_size`` is
            ``3``, prefer::

                void Foo() {
                ...if (isBar1
                .........&& isBar2)
                ......bar();
                }

        :param require_braces_at_namespace:
            Attach braces to a namespace statement. This is done
            regardless of the brace style being used.
            For example: If set to ``True``, prefer::

                namespace FooName {
                ...
                }

        :param require_braces_at_class:
            Attach braces to a class statement. This is done regardless of the
            brace style being used.
            For example: If set to ``True``, prefer::

                class FooClass {
                ...
                };

        :param require_braces_at_inline:
            Attach braces to class and struct inline function definitions. This
            option has precedence for all styles except ``Horstmann`` and
            ``Pico`` (run-in styles). It is effective for C++ files only.
            For example: If set to ``True``, prefer::

                class FooClass
                {
                    void Foo() {
                    ...
                    }
                };

        :param require_braces_at_extern:
            Attach braces to a braced extern "C" statement. This is done
            regardless of the brace style being used. This option is effective
            for C++ files only.
            For example: If set to ``True``, prefer::

                #ifdef __cplusplus
                extern "C" {
                #endif

        :param allow_indent_classes:
            Indent ``class`` and ``struct`` blocks so that the entire block is
            indented. The ``struct`` blocks are indented only if an access
            modifier, ``public:``, ``protected:`` or ``private:``, is declared
            somewhere in the ``struct``. This option is effective for C++ files
            only. For example: If set to ``True``, prefer this::

                class Foo
                {
                    public:
                        Foo();
                        virtual ~Foo();
                };

            over this::

                class Foo
                {
                public:
                    Foo();
                    virtual ~Foo();
                };

        :param allow_indent_modifiers:
            Indent ``class`` and ``struct`` access modifiers, ``public:``,
            ``protected:`` and ``private:``, one half indent. The rest of the
            class is not indented. This option is effective for C++ files only.
            For example: If set to ``True``, prefer this::

                class Foo
                {
                  public:
                    Foo();
                    virtual ~Foo();
                };

            over this::

                class Foo
                {
                public:
                    Foo();
                    virtual ~Foo();
                };

        :param allow_indent_switches:
            Indent ``switch`` blocks so that the ``case X:`` statements are
            indented in the switch block. The entire case block is indented.
            For example: If set to ``True``, prefer this::

                switch (foo)
                {
                    case 1:
                        a += 1;
                        break;

                    case 2:
                    {
                        a += 2;
                        break;
                    }
                }

            over this::

                switch (foo)
                {
                case 1:
                    a += 1;
                    break;

                case 2:
                {
                    a += 2;
                    break;
                }
                }

        :param allow_indent_cases:
            Indent ``case X:`` blocks from the ``case X:`` headers. Case
            statements not enclosed in blocks are NOT indented.
            For example: If set to ``True``, prefer this::

                switch (foo)
                {
                    case 1:
                        a += 1;
                        break;

                    case 2:
                        {
                            a += 2;
                            break;
                        }
                }

            over this::

                switch (foo)
                {
                    case 1:
                        a += 1;
                        break;

                    case 2:
                    {
                        a += 2;
                        break;
                    }
                }

        :param allow_indent_namespaces:
            Add extra indentation to namespace blocks. This option has no
            effect on Java files.
            For example: If set to ``True``, prefer this::

                namespace foospace
                {
                    class Foo
                    {
                        public:
                            Foo();
                            virtual ~Foo();
                    };
                }

            over this::

                namespace foospace
                {
                class Foo
                {
                    public:
                        Foo();
                        virtual ~Foo();
                };
                }

        :param allow_indent_labels:
            Add extra indentation to labels so they appear 1 indent less than
            the current indentation, rather than being flushed to the
            left (the default).
            For example: If set to ``True``, prefer this::

                void Foo() {
                    while (isFoo) {
                        if (isFoo)
                            goto error;
                        ...
                    error:
                        ...
                        }
                }

            over this::

                void Foo() {
                    while (isFoo) {
                        if (isFoo)
                            goto error;
                        ...
                error:
                        ...
                        }
                }

        :param allow_indent_preproc_block:
            Indent preprocessor blocks at brace level zero and immediately
            within a namespace. There are restrictions on what will be
            indented. Blocks within methods, classes, arrays, etc., will not
            be indented. Blocks containing braces or multi-line define
            statements will not be indented. Without this option the
            preprocessor block is not indented.
            For example: If set to ``True``, prefer this::

                #ifdef _WIN32
                    #include <windows.h>
                    #ifndef NO_EXPORT
                        #define EXPORT
                    #endif
                #endif

            over this::

                #ifdef _WIN32
                #include <windows.h>
                #ifndef NO_EXPORT
                #define EXPORT
                #endif
                #endif

        :param allow_indent_preproc_definition:
            Indent multi-line preprocessor definitions ending with a backslash.
            Should be used with ``convert_tabs_to_spaces`` for proper results.
            Does a pretty good job, but cannot perform miracles in obfuscated
            preprocessor definitions. Without this option the preprocessor
            statements remain unchanged.
            For example: If set to ``True``, prefer this::

                #define Is_Bar(arg,a,b) \
                    (Is_Foo((arg), (a)) \
                     || Is_Foo((arg), (b)))

            over this::

                #define Is_Bar(arg,a,b) \
                (Is_Foo((arg), (a)) \
                || Is_Foo((arg), (b)))

        :param allow_indent_preproc_conditionals:
            Indent preprocessor conditional statements to the same level as the
            source code.
            For example: If set to ``True``, prefer this::

                        isFoo = true;
                        #ifdef UNICODE
                        text = wideBuff;
                        #else
                        text = buff;
                        #endif
            over this::

                        isFoo = true;
                #ifdef UNICODE
                        text = wideBuff;
                #else
                        text = buff;
                #endif

        :param allow_indent_column_one_comments:
            Indent C++ comments beginning in column one. By default C++
            comments beginning in column one are assumed to be commented-out
            code and not indented. This option will allow the comments to be
            indented with the code.
            For example: If set to ``True``, prefer this::

                void Foo()\n"
                {
                    // comment
                    if (isFoo)
                        bar();
                }

            over this::

                void Foo()\n"
                {
                // comment
                    if (isFoo)
                        bar();
                }

        :param allow_pad_header_blocks:
            Pad empty lines around header blocks
            (e.g. ``if``, ``for``, ``while``...).
            For example: If set to ``True``, prefer this::

                isFoo = true;

                if (isFoo) {
                    bar();
                } else {
                    anotherBar();
                }

                isBar = false;

            over this::

                isFoo = true;
                if (isFoo) {
                    bar();
                } else {
                    anotherBar();
                }
                isBar = false;

        :param allow_pad_operators:
            Insert space padding around operators. This will also pad commas.
            For example: If set to ``True``, prefer this::

                if (foo == 2)
                    a = bar((b - c) * a, d--);

            over this::

                if (foo==2)
                    a=bar((b-c)*a,d--);

        :param allow_pad_parenthesis:
            Insert space padding around parenthesis on both the outside and the
            inside.
            For example: If set to ``True``, prefer this::

                if ( isFoo ( ( a+2 ), b ) )
                    bar ( a, b );

            over this::

                if (isFoo((a+2), b))
                    bar(a, b);

        :param allow_pad_parenthesis_out:
            Insert space padding around parenthesis on the outside only.
            Parenthesis that are empty will not be padded.
            For example: If set to ``True``, prefer this::

                if (isFoo ( (a+2), b) )
                    bar (a, b);

            over this::

                if (isFoo((a+2), b))
                    bar(a, b);

        :param allow_pad_parenthesis_in:
            Insert space padding around parenthesis on the inside only.
            For example: If set to ``True``, prefer this::

                if ( isFoo( ( a+2 ), b ) )
                    bar( a, b );

            over this::

                if (isFoo((a+2), b))
                    bar(a, b);

        :param prohibit_empty_lines_in_func:
            Delete empty lines within a function or method. Empty lines outside
            of functions or methods are NOT deleted.
            For example: If set to ``True``, prefer this::

                void Foo()
                {
                    foo1 = 1;
                    foo2 = 2;
                }

            over this::

                void Foo()
                {

                    foo1 = 1;

                    foo2 = 2;

                }

        :param break_closing_braces:
            When used with some specific ``bracket_style``, this breaks closing
            headers (e.g. ``else``, ``catch``, ...) from their immediately
            preceding closing braces. Closing header braces are always broken
            with the other styles.
            For example: If set to ``True``, prefer this::

                void Foo(bool isFoo) {
                    if (isFoo) {
                        bar();
                    }
                    else {
                        anotherBar();
                    }
                }

            over this::

                void Foo(bool isFoo) {
                    if (isFoo) {
                        bar();
                    } else {
                        anotherBar();
                    }
                }

        :param break_elseifs:
            Break ``else if`` header combinations into separate lines.
            For example: If set to ``True``, prefer this::

                if (isFoo) {
                    bar();
                }
                else
                    if (isFoo1()) {
                        bar1();
                    }
                    else
                        if (isFoo2()) {
                            bar2();
                        }

            over this::

                if (isFoo) {
                    bar();
                }
                else if (isFoo1()) {
                    bar1();
                }
                else if (isFoo2()) {
                    bar2;
                }

        :param break_one_line_headers:
            Break one line headers (e.g. ``if``, ``while``, ``else``, ...) from
            a statement residing on the same line. If the statement is enclosed
            in braces, the braces will be formatted according to the requested
            brace style.
            For example: If set to ``True``, prefer this::

                void Foo(bool isFoo)
                {
                    if (isFoo1)
                        bar1();

                    if (isFoo2) {
                        bar2();
                    }
                }

            over this::

                void Foo(bool isFoo)
                {
                    if (isFoo1) bar1();

                    if (isFoo2) { bar2(); }
                }

        :param require_braces_at_one_line_conditionals:
            Add braces to unbraced one line conditional statements
            (e.g. ``if``, ``for``, ``while``...). The statement must be on a
            single line. The braces will be added according to the requested
            brace style.
            For example: If set to ``True``, prefer this::

                if (isFoo) {
                    isFoo = false;
                }

            over this::

                if (isFoo)
                    isFoo = false;

        :param prohibit_braces_from_one_line_conditionals:
            Remove braces from conditional statements
            (e.g. ``if``, ``for``, ``while``...). The statement must be a
            single statement on a single line.
            For example: If set to ``True``, prefer this::

                if (isFoo)
                    isFoo = false;

            over this::

                if (isFoo)
                {
                    isFoo = false;
                }

        :param prohibit_comment_prefix:
            Remove the preceding '*' in a multi-line comment that begins a
            line. A trailing '*', if present, is also removed. Text that is
            less than one indent is indented to one indent. Text greater than
            one indent is not changed. Multi-line comments that begin a line,
            but without the preceding '*', are indented to one indent for
            consistency. This can slightly modify the indentation of commented
            out blocks of code. Lines containing all '*' are left unchanged.
            Extra spacing is removed from the comment close '*/'.
            For example: If set to ``True``, prefer this::

                /*
                    comment line 1
                    comment line 2
                */

            over this::

                /*
                 * comment line 1
                 * comment line 2
                 */

        """
        rules_map = {
            '--attach-namespaces': require_braces_at_namespace,
            '--attach-classes': require_braces_at_class,
            '--attach-inlines': require_braces_at_inline,
            '--attach-extern-c': require_braces_at_extern,
            '--indent-classes': allow_indent_classes,
            '--indent-modifiers': allow_indent_modifiers,
            '--indent-switches': allow_indent_switches,
            '--indent-cases': allow_indent_cases,
            '--indent-namespaces': allow_indent_namespaces,
            '--indent-labels': allow_indent_labels,
            '--indent-preproc-block': allow_indent_preproc_block,
            '--indent-preproc-define': allow_indent_preproc_definition,
            '--indent-preproc-cond': allow_indent_preproc_conditionals,
            '--indent-col1-comments': allow_indent_column_one_comments,
            '--break-blocks': allow_pad_header_blocks,
            '--pad-oper': allow_pad_operators,
            '--pad-paren': allow_pad_parenthesis,
            '--pad-paren-out': allow_pad_parenthesis_out,
            '--pad-paren-in': allow_pad_parenthesis_in,
            '--delete-empty-lines': prohibit_empty_lines_in_func,
            '--break-closing-brackets': break_closing_braces,
            '--break-elseifs': break_elseifs,
            '--break-one-line-headers': break_one_line_headers,
            '--add-brackets': require_braces_at_one_line_conditionals,
            '--remove-brackets': prohibit_braces_from_one_line_conditionals,
            '--remove-comment-prefix': prohibit_comment_prefix
        }
        args = ['--suffix=none', '--dry-run']
        if bracket_style:
            args.append('--style=' + bracket_style)
        if use_spaces is True:
            args.append('-s' + str(indent_size))
        elif use_spaces is False:
            args.append('-t' + str(indent_size))
        args += (k for k, v in rules_map.items() if v)
        return args



import shutil

from coalib.bearlib.languages.documentation.DocumentationComment import (
    DocumentationComment)
from coalib.bearlib.languages.documentation.DocstyleDefinition import (
    DocstyleDefinition)
from coalib.bearlib.languages.documentation.DocBaseClass import (
    DocBaseClass)
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result
from coalib.settings.Setting import typed_list


class DocGrammarBear(DocBaseClass, LocalBear):
    LANGUAGES = {language for docstyle, language in
                 DocstyleDefinition.get_available_definitions()}
    REQUIREMENTS = {PipRequirement('language-check', '1.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/132564'
    CAN_FIX = {'Documentation', 'Spelling', 'Grammar'}

    @classmethod
    def check_prerequisites(cls):
        if shutil.which('java') is None:
            return 'java is not installed.'
        else:
            try:
                from language_check import LanguageTool, correct
                LanguageTool
                correct
                return True
            except ImportError:  # pragma: no cover
                return 'Please install the `language-check` pip package.'

    def process_documentation(self,
                              parsed,
                              locale,
                              languagetool_disable_rules):
        """
        This fixes the parsed documentation comment by applying spell checking
        and grammatic rules via LanguageTool.

        :param parsed:
            Contains parsed documentation comment.
        :param locale:
            A locale representing the language you want to have checked.
            Default is set to 'en-US'.
        :param languagetool_disable_rules:
            List of rules to disable checks for.
        :return:
            A tuple of fixed parsed documentation comment and warning_desc.
        """
        # Defer import so the check_prerequisites can be run without
        # language_check being there.
        from language_check import LanguageTool, correct

        tool = LanguageTool(locale)
        tool.disabled.update(languagetool_disable_rules)

        metadata = iter(parsed)

        new_metadata = []
        for comment in metadata:
            matches = tool.check(comment.desc)
            new_desc = correct(comment.desc, matches)
            new_metadata.append(comment._replace(desc=new_desc))

        return (new_metadata,
                'Documentation has invalid Grammar/Spelling')

    def run(self, filename, file, language: str,
            docstyle: str='default', locale: str='en-US',
            languagetool_disable_rules: typed_list(str)=()):
        """
        Checks the main description and comments description of documentation
        with LanguageTool. LanguageTool finds many errors that a simple spell
        checker cannot detect and several grammar problems. A full list of
        rules for english language can be found here at:
        https://community.languagetool.org/rule/list?lang=en
        LanguageTool currently supports more than 25 languages. For further
        information, visit: https://www.languagetool.org/languages/

        :param language:
            The programming language of the file(s).
        :param docstyle:
            The docstyle to use. For example ``default`` or ``doxygen``.
            Docstyles are language dependent, meaning not every language is
            supported by a certain docstyle.
        :param locale:
            A locale representing the language you want to have checked.
            Default is set to 'en-US'.
        :param languagetool_disable_rules:
            List of rules to disable checks for.
        """
        for doc_comment in self.extract(file, language, docstyle):
            parsed = doc_comment.parse()

            (new_metadata, warning_desc) = self.process_documentation(
                parsed, locale, languagetool_disable_rules)

            new_comment = DocumentationComment.from_metadata(
                new_metadata, doc_comment.docstyle_definition,
                doc_comment.marker, doc_comment.indent, doc_comment.position)

            if new_comment != doc_comment:
                # Something changed, let's apply a result.
                diff = self.generate_diff(file, doc_comment, new_comment)

                yield Result(
                    origin=self,
                    message=warning_desc,
                    affected_code=(diff.range(filename),),
                    diffs={filename: diff})



from coalib.bearlib.abstractions.Linter import linter
from coalib.settings.Setting import typed_list
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY

from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='httpolice',
        output_format='regex',
        output_regex=r'(?P<severity>E|C|D)\s(?P<message>.*)',
        severity_map={'E': RESULT_SEVERITY.MAJOR,
                      'C': RESULT_SEVERITY.NORMAL,
                      'D': RESULT_SEVERITY.INFO})
class HTTPoliceLintBear:
    """
    HTTPolice is a linter for HTTP requests and responses. It checks them for
    conformance to standards and best practices.
    """
    LANGUAGES = {'HAR'}
    REQUIREMENTS = {PipRequirement('HTTPolice', '0.5.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax'}
    SEE_MORE = 'https://github.com/vfaronov/httpolice'

    @staticmethod
    def create_arguments(filename, file, config_file,
                         httpolice_silence_ids: typed_list(str)=[]):
        """
        :param httpolice_silence_ids:
            Silences the given list of notice IDs. You can get more information
            about the available notices and their IDs at
            https://httpolice.readthedocs.io/en/stable/notices.html.
        """
        args = '-i', 'har', filename
        if httpolice_silence_ids:
            args += tuple('-s=' + part for part in httpolice_silence_ids)
        return args



import json

from coalib.bearlib.abstractions.Linter import linter
from coalib.settings.Setting import typed_list

from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='pug-lint',
        output_format='regex',
        output_regex=r'(?P<line>\d+):?(?P<column>\d+)? (?P<message>.+)',
        use_stdout=False,
        use_stderr=True)
class PugLintBear:
    """
    A configurable linter and style checker for ``Pug`` (formerly ``Jade``)
    that is a clean, whitespace-sensitive template language for writing HTML.
    """

    LANGUAGES = {'Pug'}
    REQUIREMENTS = {NpmRequirement('pug-lint', '2.4.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax', 'Redundancy'}
    SEE_MORE = 'https://github.com/pugjs/pug-lint'

    @staticmethod
    def generate_config(filename, file,
                        prohibit_block_expansion: bool=True,
                        prohibit_class_attribute_with_static_value: bool=True,
                        prohibit_class_literals_before_attributes: bool=True,
                        prohibit_class_literals_before_id_literals: bool=True,
                        prohibit_class_literals: bool=True,
                        prohibit_duplicate_attributes: bool=True,
                        prohibit_html_text: bool=True,
                        prohibit_id_attribute_with_static_value: bool=True,
                        prohibit_id_literals_before_attributes: bool=True,
                        prohibit_id_literals: bool=True,
                        prohibit_legacy_mixin_call: bool=True,
                        prohibit_multiple_line_breaks: bool=False,
                        prohibit_spaces_inside_attribute_brackets: bool=True,
                        prohibit_string_interpolation: bool=True,
                        prohibit_tag_interpolation: bool=True,
                        prohibit_specific_attributes: typed_list(str)=None,
                        prohibit_specific_tags: typed_list(str)=None,
                        enforce_class_literals_before_attributes: bool=False,
                        enforce_class_literals_before_id_literals: bool=False,
                        enforce_id_literals_before_attributes: bool=False,
                        enforce_lower_case_attributes: bool=True,
                        enforce_lower_case_tags: bool=True,
                        enforce_spaces_inside_attribute_brackets: bool=False,
                        enforce_strict_equality_operators: bool=True,
                        validate_div_tags: bool=True,
                        validate_extensions: bool=True,
                        validate_self_closing_tags: bool=True,
                        preferred_quotation: str="'",
                        max_lines_per_file: int=None,
                        puglint_config: str=''):
        """
        :param prohibit_block_expansion:
            When ``True``, disallow any block expansion operators.
            For example: If set to ``True``, this will throw a warning::

                p: strong text
                table: tr: td text

        :param prohibit_class_attribute_with_static_value:
            When ``True``, prefer class literals over class attributes with
            static values.
            For example: If set to ``True``, prefer ``span.foo`` over
            ``span(class='foo')``.
        :param prohibit_class_literals_before_attributes:
            When ``True``, prefer all attribute blocks to be written before
            any class literals.
            For example: If set to ``True``, prefer
            ``input(type='text').class`` over ``input.class(type='text')``.
        :param prohibit_class_literals_before_id_literals:
            When ``True``, prefer all ID literals to be written before any
            class literals.
            For example: If set to ``True``, prefer
            ``input#id.class(type='text')`` over
            ``input.class#id(type='text')``.
        :param prohibit_class_literals:
            When ``True``, disallow any class literals.
            For example: If set to ``True``, prefer ``div(class='class')``
            over ``.class``.
        :param prohibit_duplicate_attributes:
            When ``True``, attribute blocks should not contain any duplicates.
            For example: If set to ``True``, this will throw a warning::

                div(a='a' a='b')
                #id(id='id')

        :param prohibit_html_text:
            When ``True``, disallow any HTML text.
            For example: If set to ``True``, this will throw a warning::

                <strong>html text</strong>
                p this is <strong>html</strong> text

        :param prohibit_id_attribute_with_static_value:
            When ``True``, prefer ID literals over ``id`` attributes with
            static values.
            For example: If set to ``True``, prefer ``span#id`` over
            ``span(id='foo')``.
        :param prohibit_id_literals_before_attributes:
            When ``True``, prefer all attribute blocks to be written before
            any ID literals.
            For example: If set to ``True``, prefer ``input(type='text')#id``
            over ``input#id(type='text')``.
        :param prohibit_id_literals:
            When ``True``, disallow any ID literals.
            For example: If set to ``True``, ``#id`` will throw a warning.
        :param prohibit_legacy_mixin_call:
            When ``True``, disallow any legacy mixin call.
            When ``True``, prefer ``+myMixin(arg)`` over
            ``mixin myMixin(arg)``.
        :param prohibit_multiple_line_breaks:
            When ``True``, disallow multiple blank lines in a row.
        :param prohibit_spaces_inside_attribute_brackets:
            When ``True``, disallow space after opening attribute bracket and
            before closing attribute bracket.
            For example: If set to ``True``, prefer
            ``input(type='text' name='name' value='value')`` over
            ``input( type='text' name='name' value='value' )``.
        :param prohibit_string_interpolation:
            When ``True``, disallow any string interpolation operators.
            For example: If set to ``True``, ``h1 #{title} text`` will throw
            a warning.
        :param prohibit_tag_interpolation:
            When ``True``, disallow any tag interpolation operators.
            For example: If set to ``True``, this will throw a warning::

                | #[strong html] text
                p #[strong html] text

        :param prohibit_specific_attributes:
            Disallow any of the attributes specified.
        :param prohibit_specific_tags:
            Disallow any of the tags specified.
        :param enforce_class_literals_before_attributes:
            When ``True``, all class literals must be written before any
            attribute blocks.
        :param enforce_class_literals_before_id_literals:
            When ``True``, all class literals should be written before any
            ID literals.
        :param enforce_id_literals_before_attributes:
            When ``True``, all ID literals must be written before any
            attribute blocks.
        :param enforce_lower_case_attributes:
            When ``True``, all attributes should be written in lower case.
            For example: If set to ``True``, prefer ``div(class='class')``
            over ``div(Class='class')``.
        :param enforce_lower_case_tags:
            When ``True``, all tags must be written in lower case.
            For example: If set to ``True``, prefer ``div(class='class')``
            over ``Div(class='class')``.
        :param enforce_spaces_inside_attribute_brackets:
            When ``True``, enforce space after opening attribute bracket and
            before closing attribute bracket.
        :param enforce_strict_equality_operators:
            When ``True``, enforce the use of ``===`` and ``!==`` instead of
            ``==`` and ``!=``.
        :param validate_div_tags:
            When ``True``, disallow any unnecessary ``div`` tags.
        :param validate_extensions:
            When ``True``, enforce proper file extensions with inclusion and
            inheritance.
        :param validate_self_closing_tags:
            When ``True``, disallow any unnecessary self closing tags.
        :param preferred_quotation:
            Your preferred quotation character, e.g.``"`` or ``'``.
        :param max_lines_per_file:
            Number of lines allowed per file.
        """
        if puglint_config:
            return None
        else:
            options = {
                'disallowBlockExpansion': prohibit_block_expansion,
                'disallowClassAttributeWithStaticValue':
                    prohibit_class_attribute_with_static_value,
                'disallowClassLiteralsBeforeAttributes':
                    prohibit_class_literals_before_attributes,
                'disallowClassLiteralsBeforeIdLiterals':
                    prohibit_class_literals_before_id_literals,
                'disallowClassLiterals': prohibit_class_literals,
                'disallowDuplicateAttributes': prohibit_duplicate_attributes,
                'disallowHtmlText': prohibit_html_text,
                'disallowIdAttributeWithStaticValue':
                    prohibit_id_attribute_with_static_value,
                'disallowIdLiteralsBeforeAttributes':
                    prohibit_id_literals_before_attributes,
                'disallowIdLiterals': prohibit_id_literals,
                'disallowLegacyMixinCall': prohibit_legacy_mixin_call,
                'disallowMultipleLineBreaks': prohibit_multiple_line_breaks,
                'disallowSpacesInsideAttributeBrackets':
                    prohibit_spaces_inside_attribute_brackets,
                'disallowStringInterpolation': prohibit_string_interpolation,
                'disallowTagInterpolation': prohibit_tag_interpolation,
                'disallowSpecificAttributes': prohibit_specific_attributes,
                'disallowSpecificTags': prohibit_specific_tags,
                'requireClassLiteralsBeforeAttributes':
                    enforce_class_literals_before_attributes,
                'requireClassLiteralsBeforeIdLiterals':
                    enforce_class_literals_before_id_literals,
                'requireIdLiteralsBeforeAttributes':
                    enforce_id_literals_before_attributes,
                'requireLowerCaseAttributes': enforce_lower_case_attributes,
                'requireLowerCaseTags': enforce_lower_case_tags,
                'requireSpacesInsideAttributeBrackets':
                    enforce_spaces_inside_attribute_brackets,
                'requireStrictEqualityOperators':
                    enforce_strict_equality_operators,
                'validateDivTags': validate_div_tags,
                'validateExtensions': validate_extensions,
                'validateSelfClosingTags': validate_self_closing_tags,
                'validateAttributeQuoteMarks': preferred_quotation,
                'maximumNumberOfLines': max_lines_per_file
            }

            for k, v in options.items():
                options[k] = v if v else None

            return json.dumps(options)

    @staticmethod
    def create_arguments(filename, file, config_file, puglint_config: str=''):
        """
        :param puglint_config:
            The location of a custom ``.pug-lintrc`` config file.
        """
        return ('--config',
                puglint_config if puglint_config else config_file,
                '--reporter', 'inline', filename)



import autoflake

from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.bearlib.aspects import map_setting_to_aspect
from coalib.bearlib.aspects.Redundancy import (
    Redundancy,
    UnusedImport,
    UnusedLocalVariable,
)


class PyUnusedCodeBear(
        LocalBear,
        aspects={
            'fix': [
                UnusedImport,
                UnusedLocalVariable,
            ]},
        languages=['Python'],
        ):
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('autoflake', '0.7')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Unused Code'}

    @map_setting_to_aspect(
        remove_all_unused_imports=UnusedImport.remove_non_standard_import,
        remove_unused_variables=UnusedLocalVariable)
    def run(self, filename, file,
            remove_all_unused_imports: bool=True,
            remove_unused_variables: bool=True):
        """
        Detects unused code. By default this functionality is limited to:

        - Unneeded pass statements.
        - All unused imports - might have side effects
        - Unused variables - might have side effects

        :param remove_all_unused_imports:
            ``False`` removes only unused builtin imports
        :param remove_unused_variables:
            ``False`` keeps unused variables
        """
        corrected = autoflake.fix_code(
                       ''.join(file),
                       additional_imports=None,
                       remove_all_unused_imports=remove_all_unused_imports,
                       remove_unused_variables=remove_unused_variables
                       ).splitlines(True)

        for diff in Diff.from_string_arrays(file, corrected).split_diff():
            yield Result(self,
                         'This file contains unused source code.',
                         affected_code=(diff.range(filename),),
                         diffs={filename: diff},
                         aspect=Redundancy('py'))



import re
import requests
from urllib.parse import urlparse
from aenum import Flag

from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.bearlib import deprecate_settings
from coalib.settings.Setting import typed_list
from coalib.results.HiddenResult import HiddenResult
from coalib.results.Result import Result
from coalib.results.SourceRange import SourceRange
from coalib.parsing.Globbing import fnmatch
from coalib.settings.Setting import typed_dict
from coala_utils.decorators import (enforce_signature, generate_ordering,
                                    generate_repr)


class LINK_CONTEXT(Flag):
    no_context = 0
    xml_namespace = 1
    pip_vcs_url = 2


@generate_repr(('id', hex),
               'origin',
               'affected_code',
               'message',
               'link',
               'http_status_code',
               'link_context')
@generate_ordering('affected_code',
                   'link',
                   'http_status_code',
                   'link_context',
                   'contents',
                   'severity',
                   'confidence',
                   'origin',
                   'message_base',
                   'message_arguments',
                   'aspect',
                   'additional_info',
                   'debug_msg')
class URLResult(HiddenResult):

    @enforce_signature
    def __init__(self, origin, affected_code,
                 link: str,
                 http_status_code: (int, None),
                 link_context: LINK_CONTEXT):

        Result.__init__(self, origin,
                        '%s responds with HTTP %s' % (link, http_status_code),
                        affected_code)

        self.contents = [affected_code[0].start.line, link, http_status_code,
                         link_context]
        self.link = link
        self.http_status_code = http_status_code
        self.link_context = link_context


class URLBear(LocalBear):
    DEFAULT_TIMEOUT = 15
    LANGUAGES = {'All'}
    REQUIREMENTS = {PipRequirement('requests', '2.12'),
                    PipRequirement('aenum', '2.0.8')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Documentation'}

    # IP Address of www.google.com
    check_connection_url = 'http://216.58.218.174'

    @classmethod
    def check_prerequisites(cls):
        code = cls.get_status_code(
            cls.check_connection_url, cls.DEFAULT_TIMEOUT)
        return ('You are not connected to the internet.'
                if code is None else True)

    @staticmethod
    def get_status_code(url, timeout):
        try:
            code = requests.head(url, allow_redirects=False,
                                 timeout=timeout).status_code
            return code
        except requests.exceptions.RequestException:
            pass

    @staticmethod
    def parse_pip_vcs_url(link):
        splitted_at = link.split('@')[0]
        splitted_schema = splitted_at[splitted_at.index('+') + 1:]
        return splitted_schema

    @staticmethod
    def extract_links_from_file(file, link_ignore_regex, link_ignore_list):
        link_ignore_regex = re.compile(link_ignore_regex)
        regex = re.compile(
            r"""
            ((git\+|bzr\+|svn\+|hg\+|)  # For VCS URLs
            https?://                   # http:// or https:// as only these
                                        # are supported by the ``requests``
                                        # library
            [^.:%\s_/?#[\]@\\]+         # Initial part of domain
            \.                          # A required dot `.`
            (
                ((?:%[A-Fa-f0-9][A-Fa-f0-9])*[^\s()%\'"`<>|\\\[\]]+)
                                        # Path name
                                        # This part allows precentage
                                        # encoding like %3F
                                        # and does not allow
                                        # any parenthesis: balanced or
                                        # unbalanced.
            |                           # OR
                \((?:%[A-Fa-f0-9][A-Fa-f0-9])*[^\s()%\'"`<>|\\\[\]]*\)
                                        # Path name contained within ()
                                        # This part allows path names that
                                        # are explicitly enclosed within one
                                        # set of parenthesis.
                                        # An example can be:
                                        # http://wik.org/Hello_(Adele_song)/200
            )
            *)
                                        # Thus, the whole part above
                                        # prevents matching of
                                        # Unbalanced parenthesis
            (?<!\.)(?<!,)               # Exclude trailing `.` or `,` from URL
            """, re.VERBOSE)
        file_context = {}
        for line_number, line in enumerate(file):
            xmlns_regex = re.compile(r'xmlns:?\w*="(.*)"')
            for match in re.findall(regex, line):
                link = match[0]
                link_context = file_context.get(link)
                if not link_context:
                    link_context = LINK_CONTEXT.no_context
                    xmlns_match = xmlns_regex.search(line)
                    if xmlns_match and link in xmlns_match.groups():
                        link_context |= LINK_CONTEXT.xml_namespace
                    if link.startswith(('hg+', 'bzr+', 'git+', 'svn+')):
                        link_context |= LINK_CONTEXT.pip_vcs_url
                    file_context[link] = link_context
                if not (link_ignore_regex.search(link) or
                        fnmatch(link, link_ignore_list)):
                    yield link, line_number, link_context

    def analyze_links_in_file(self, file, network_timeout, link_ignore_regex,
                              link_ignore_list):
        for link, line_number, link_context in self.extract_links_from_file(
                file, link_ignore_regex, link_ignore_list):

            if link_context is link_context.pip_vcs_url:
                link = URLBear.parse_pip_vcs_url(link)

            host = urlparse(link).netloc
            code = URLBear.get_status_code(
                link,
                network_timeout.get(host)
                if host in network_timeout
                else network_timeout.get('*')
                if '*' in network_timeout
                else URLBear.DEFAULT_TIMEOUT)
            yield line_number + 1, link, code, link_context

    @deprecate_settings(link_ignore_regex='ignore_regex',
                        network_timeout=('timeout', lambda t: {'*': t}))
    def run(self, filename, file,
            network_timeout: typed_dict(str, int, DEFAULT_TIMEOUT)=dict(),
            link_ignore_regex: str='([.\/]example\.com|\{|\$)',
            link_ignore_list: typed_list(str)=''):
        """
        Find links in any text file.

        Warning: This bear will make HEAD requests to all URLs mentioned in
        your codebase, which can potentially be destructive. As an example,
        this bear would naively just visit the URL from a line that goes like
        `do_not_ever_open = 'https://api.acme.inc/delete-all-data'` wiping out
        all your data.

        :param network_timeout:       A dict mapping URLs and timeout to be
                                      used for that URL. All the URLs that have
                                      the same host as that of URLs provided
                                      will be passed that timeout. It can also
                                      contain a wildcard timeout entry with key
                                      '*'. The timeout of all the websites not
                                      in the dict will be the value of the key
                                      '*'.
        :param link_ignore_regex:     A regex for urls to ignore.
        :param link_ignore_list: Comma separated url globs to ignore
        """
        network_timeout = {urlparse(url).netloc
                           if not url == '*' else '*': timeout
                           for url, timeout in network_timeout.items()}

        for line_number, link, code, context in self.analyze_links_in_file(
                file, network_timeout, link_ignore_regex, link_ignore_list):
            affected_code = SourceRange.from_values(filename, line_number)

            yield URLResult(self, (affected_code,), link, code, context)



import json

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement

_setting_map = {True: 'always',
                False: 'never',
                None: 'false'}


@linter(executable='stylint',
        output_format='regex',
        output_regex=r'(?P<line>\d+):?(?P<column>\d+)?\s+.*?'
                     r'(?P<severity>error|warning)\s+(?P<message>.+?)'
                     r'(?:  .*|\n|$)')
class StylintBear:
    """
    Attempts to catch little mistakes (duplication of rules for instance) and
    enforces a code style guide on Stylus (a dynamic stylesheet language
    with the ``.styl`` extension that is compiled into CSS) files.

    The ``StylintBear`` is able to catch following problems:
    - Duplication of rules
    - Mixed spaces and tabs
    - Unnecessary brackets
    - Missing colon between property and value
    - Naming conventions
    - Trailing whitespace
    - Consistent quotation style
    - Use of extra spaces inside parenthesis
    - Naming convention when declaring classes, ids, and variables
    - Unnecessary leading zeroes on decimal points
    - Checks if a property is valid CSS or HTML
    """

    LANGUAGES = {'Stylus'}
    REQUIREMENTS = {NpmRequirement('stylint', '1.5.9')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax', 'Redundancy'}
    SEE_MORE = 'https://github.com/SimenB/stylint'

    @staticmethod
    def generate_config(filename, file,
                        block_keyword: bool=None,
                        brackets: bool=False,
                        colons_for_property_declaration: bool=True,
                        color_variables_for_hex_values: bool=True,
                        spaces_after_commas: bool=True,
                        spaces_after_comments: bool=True,
                        allow_trailing_whitespace: bool=False,
                        no_css_literals: bool=False,
                        max_selector_depth: bool=False,
                        check_duplicates: bool=True,
                        efficient_properties: bool=True,
                        extend_preference: str=None,
                        indent_size: int=0,
                        leading_zero: bool=None,
                        max_errors: int=0,
                        max_warnings: int=0,
                        mixed_spaces_and_tabs: bool=False,
                        variable_naming_convention: str=None,
                        strict_naming_convention: bool=False,
                        none_keyword: bool=False,
                        check_no_important_keyword: bool=True,
                        spaces_inside_parentheses: bool=None,
                        placeholder: bool=True,
                        prefix_vars_with_dollar: bool=True,
                        semicolons: bool=False,
                        sort_order: str='alphabetical',
                        stacked_properties: bool=False,
                        check_property_validity: bool=True,
                        preferred_quotation: str=None,
                        zero_units: bool=False,
                        z_index_normalize_base: int=0,
                        stylint_config: str=''):
        """
        :param block_keyword:
            When ``True`` expect the ``@block`` keyword when defining block
            variables. When ``False``, expect no ``@block`` keyword when
            defining block variables.
        :param brackets:
            When ``True``, expect ``{}`` when declaring a selector. When
            ``False``, expect no brackets when declaring a selector.
        :param colons_for_property_declaration:
            When ``True``, expect ``:`` when declaring a property. When
            ``False``, expect no ``:`` when declaring a property.
        :param color_variables_for_hex_values:
            When ``True``, enforce variables when defining hex values.
        :param spaces_after_commas:
            Enforce or disallow spaces after commas.
        :param spaces_after_comments:
            Enforce or disallow spaces after line comments.
            For example: If set to ``True``, prefer
            ``// comment`` over ``//comment``.
        :param allow_trailing_whitespace:
            If ``True``, ignores trailing whitespace. If ``False``, trailing
            whitespace will throw a warning.
        :param no_css_literals:
            By default Stylint ignores ``@css`` blocks. If set to ``True``
            however, warnings are thrown if ``@css`` is used.
        :param max_selector_depth:
            Set the max selector depth. If set to 4, max selector depth will
            be 4 indents. Pseudo selectors like ``&:first-child`` or
            ``&:hover`` won't count towards the limit.
        :param check_duplicates:
            Checks if selectors or properties are duplicated unnecessarily.
        :param efficient_properties:
            Check for places where properties can be written more efficiently.
        :param extend_preference:
            Pass in either ``@extend`` or ``@extends`` and then enforce that.
            Both are valid in Stylus. It doesn't really matter which one
            you use.
            For example: If set to ``@extends``, prefer
            ``@extends $some-var`` instead of ``@extend $some-var``
            or if set to ``@extend``, prefer
            ``@extend $some-var`` over ``@extends $some-var``.
        :param indent_size:
            This works in conjunction with ``max_selector_depth``. If you
            indent with spaces this is the number of spaces you indent with.
            If you use hard tabs, set this value to ``0``.
            For example: If set to ``2``, prefer
            ``/s/smargin: 0`` over ``/s/s/smargin: 0``.
        :param leading_zero:
            When ``True``, prefer leading zeroes on decimal points. When
            ``False``, disallow leading zeroes on decimal points.
        :param max_errors:
            Set maximum number of errors. If ``0``, all errors will
            be shown without a limit.
        :param max_warnings:
            Set maximum number of warnings. If ``0``, all errors will
            be shown without a limit.
        :param mixed_spaces_and_tabs:
            If a non-negative number is passed to ``indent_size``,
            soft tabs (i.e. spaces) are assumed, and if
            ``0`` is passed to ``indent_size``, hard tabs are assumed.
            For example: If ``indent_size = 4`` and
            ``mixed_spaces_and_tabs = True``, prefer
            ``/s/s/s/smargin 0`` over ``/tmargin 0``
            or if ``indent_size = 0`` and
            ``mixed_spaces_and_tabs = True``, prefer
            ``/tmargin 0`` over ``/s/s/s/smargin 0``.
        :param variable_naming_convention:
            Enforce a particular naming convention when declaring variables.
            Throws a warning if you don't follow the convention.
            Supported values are ``lowercase-dash``, ``lowercase_underscore``,
            ``camelCase`` and ``BEM``.
        :param strict_naming_convention:
            By default, ``variable_naming_convention`` only looks at variable
            names. If ``strict_naming_convention`` is set to ``True``,
            ``variable_naming_convention`` will also look at class and
            ID names.
        :param none_keyword:
            If ``True`` check for places where ``none`` used instead of ``0``.
            If ``False`` check for places where ``0`` could be used
            instead of ``none``.
        :param check_no_important_keyword:
            If ``True``, show warning when ``!important`` is found.
            For example: If set to ``True``, this will throw a warning::

                div
                    color red !important

        :param spaces_inside_parentheses:
            Enforce or disallow use of extra spaces inside parentheses.
            For example: If set to ``True``, prefer
            ``my-mixin( $myParam )`` over ``my-mixin($myParam)``
            or if set to ``False``, prefer
            ``my-mixin($myParam)`` over ``my-mixin( $myParam )``.
        :param placeholder:
            Enforce extending placeholder vars when using ``@extend(s)``.
        :param prefix_vars_with_dollar:
            Enforce use of ``$`` when defining a variable. In Stylus, using a
            ``$`` when defining a variable is optional, but is a good idea
            if you want to prevent ambiguity. Not including the ``$`` sets up
            situations where you wonder: "Is this a variable or a value?"
            For instance: ``padding $default`` is easier to understand than
            ``padding default``.
            For example: If set to ``True``, prefer
            ``$my-var = 0`` over ``my-var = 0``
            or if set to ``False``, prefer
            ``my-var = 0`` over ``$my-var = 0``.
        :param semicolons:
            When ``True``, enforce semicolons. When ``False``, disallow
            semicolons.
        :param sort_order:
            Enforce a particular sort order when declaring properties. Throws
            a warning if you don't follow the order.
            For example: If set to ``alphabetical``, prefer this::

                .some-class
                    display block
                    float left
                    position absolute
                    right 10px
                    top 0

            over this::

                .some-class
                    position absolute
                    top 0
                    right 10px
                    display block
                    float left

            Supported values are ``alphabetical`` and ``grouped``.
        :param stacked_properties:
            No one-liners. Enforce putting properties on new lines.
            For example: If set to ``False``, prefer::

                .className
                    padding 0

            over
            ``.className { padding 0 }``
        :param check_property_validity:
            Check that a property is valid CSS or HTML.
        :param preferred_quotation:
            Your preferred quotation character, e.g. ``double`` or ``single``.
        :param zero_units:
            Looks for instances of ``0px``. You don't need the ``px``.
            Checks all units, not just ``px``.
        :param z_index_normalize_base:
            Enforce some basic z-index sanity. Any number passed in
            will be used as the base for your z-index values.
            Use ``0`` to disable this feature.
            For example: If set to ``5``, prefer
            ``z-index 10`` over ``z-index 9``
            or if set to ``10``, prefer
            ``z-index 20`` over ``z-index 15``.
        """
        if stylint_config:
            return None
        else:
            options = {
                'blocks': _setting_map[block_keyword],
                'brackets': _setting_map[brackets],
                'colons': _setting_map[colons_for_property_declaration],
                'colors': _setting_map[color_variables_for_hex_values],
                'commaSpace': _setting_map[spaces_after_commas],
                'commentSpace': _setting_map[spaces_after_comments],
                'cssLiteral': _setting_map[no_css_literals],
                'depthLimit': max_selector_depth,
                'duplicates': check_duplicates,
                'efficient': _setting_map[efficient_properties],
                'extendPref': extend_preference,
                'indentPref': indent_size,
                'leadingZero': _setting_map[leading_zero],
                'maxErrors': max_errors,
                'maxWarnings': max_warnings,
                'mixed': mixed_spaces_and_tabs,
                'namingConvention': variable_naming_convention,
                'namingConventionStrict': strict_naming_convention,
                'none': _setting_map[none_keyword],
                'noImportant': check_no_important_keyword,
                'parenSpace': _setting_map[spaces_inside_parentheses],
                'placeholders': _setting_map[placeholder],
                'prefixVarsWithDollar': _setting_map[prefix_vars_with_dollar],
                'quotePref': preferred_quotation,
                'semicolons': _setting_map[semicolons],
                'sortOrder': sort_order,
                'stackedProperties': _setting_map[stacked_properties],
                'trailingWhitespace':
                    False if allow_trailing_whitespace is True else 'never',
                'valid': check_property_validity,
                'zeroUnits': _setting_map[zero_units],
                'zIndexNormalize':
                    False if z_index_normalize_base == 0
                    else z_index_normalize_base,
                'groupOutputByFile': True,
                'reporterOptions': {
                    'columns': ['lineData', 'severity', 'description', 'rule'],
                    'columnSplitter': '  ',
                    'showHeaders': False,
                    'truncate': True
                }
            }

            return json.dumps(options)

    @staticmethod
    def create_arguments(filename, file, config_file, stylint_config: str=''):
        """
        :param stylint_config:
            The location of the ``.stylintrc`` config file.
        """
        return ('--config',
                stylint_config if stylint_config else config_file,
                filename)



import requests

from bears.general.URLBear import URLBear

from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY

from dependency_management.requirements.PipRequirement import PipRequirement

from memento_client import MementoClient


class MementoBear(LocalBear):
    DEFAULT_TIMEOUT = 15
    LANGUAGES = {'All'}
    REQUIREMENTS = {PipRequirement('memento_client', '0.6.1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Documentation'}
    BEAR_DEPS = {URLBear}

    @staticmethod
    def check_archive(mc, link):
        """
        Check the link is it archived or not.

        :param mc:   A `memento_client.MementoClient` instance.
        :param link: The link (str) that will be checked.
        :return:     Boolean, `True` means the link has been archived.
        """
        try:
            mc.get_memento_info(link)['mementos']
        except KeyError:
            return False
        return True

    @staticmethod
    def get_redirect_urls(link):
        urls = []

        resp = requests.head(link, allow_redirects=True)
        for redirect in resp.history:
            urls.append(redirect.url)

        return urls

    def run(self, filename, file, dependency_results=dict(),
            follow_redirects: bool=True):
        """
        Find links in any text file and check if they are archived.

        Link is considered valid if the link has been archived by any services
        in memento_client.

        This bear can automatically fix redirects.

        Warning: This bear will make HEAD requests to all URLs mentioned in
        your codebase, which can potentially be destructive. As an example,
        this bear would naively just visit the URL from a line that goes like
        `do_not_ever_open = 'https://api.acme.inc/delete-all-data'` wiping out
        all your data.

        :param dependency_results: Results given by URLBear.
        :param follow_redirects:   Set to true to check all redirect urls.
        """
        self._mc = MementoClient()

        for result in dependency_results.get(URLBear.name, []):
            line_number, link, code, context = result.contents

            if not (code and 200 <= code < 400):
                continue

            status = MementoBear.check_archive(self._mc, link)
            if not status:
                yield Result.from_values(
                    self,
                    ('This link is not archived yet, visit '
                     'https://web.archive.org/save/%s to get it archived.'
                     % link),
                    file=filename,
                    line=line_number,
                    severity=RESULT_SEVERITY.INFO
                )

            if follow_redirects and 300 <= code < 400:  # HTTP status 30x
                redirect_urls = MementoBear.get_redirect_urls(link)

                for url in redirect_urls:
                    status = MementoBear.check_archive(self._mc, url)
                    if not status:
                        yield Result.from_values(
                            self,
                            ('This link redirects to %s and not archived yet, '
                             'visit https://web.archive.org/save/%s to get it '
                             'archived.'
                             % (url, url)),
                            file=filename,
                            line=line_number,
                            severity=RESULT_SEVERITY.INFO
                        )



import json

from coalib.bearlib.abstractions.Linter import linter

from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='csscomb',
        output_format='corrected',
        use_stdin=True,
        result_message='The text does not comply to the set style.')
class CSSCombBear:
    """
    CSScomb is a coding style formatter for CSS. You can easily write your own
    configuration to make your style sheets beautiful and consistent.
    """

    LANGUAGES = {'CSS'}
    REQUIREMENTS = {NpmRequirement('csscomb', '4.2.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/bxpke7nizyxdf6mlx6ss4h405'
    CAN_FIX = {'Formatting'}
    SEE_MORE = 'https://github.com/csscomb/csscomb.js'

    @staticmethod
    def generate_config(filename, file,
                        enforce_semicolon: bool=True,
                        use_block_indentation: int=4,
                        use_color_case: str='upper',
                        allow_color_shorthand: bool=None,
                        allow_leading_zero_in_dimensions: bool=True,
                        preferred_quotation: str="'",
                        prohibit_empty_rulesets: bool=True,
                        use_space_after_colon: int=1,
                        use_space_before_colon: int=0,
                        use_space_after_combinator: int=1,
                        use_space_before_combinator: int=1,
                        use_space_between_declarations: int=None,
                        use_space_after_opening_brace: int=None,
                        use_space_before_closing_brace: int=None,
                        use_space_after_selector_delimiter: int=1,
                        use_space_before_selector_delimiter: int=1,
                        prohibit_trailing_whitespace: bool=True,
                        prohibit_units_in_zero_valued_dimensions: bool=True,
                        vendor_prefix_align: bool=None,
                        use_lines_between_rulesets: int=1,
                        csscomb_config: str=''):
        """
        :param enforce_semicolon:
            Whether to add a semicolon after the last value/mixin.
            For example: If set to ``True``, prefer this::

                a { color: red; text-decoration: underline; }

            over this::

                a { color: red; text-decoration: underline }

        :param use_block_indentation:
            Set the number of spaces as indent for code inside blocks,
            including media queries and nested rules.
            For example: If set to ``4``, prefer this::

                a {
                    top: 0;
                    p {
                        color: tomato;
                        position: happy;
                        }
                    }

            over this::

                a {
                top: 0;
                  p {
                      color: tomato;
                position: happy;
                 }
                }

        :param use_color_case:
            Unify case of hexadecimal colors. Acceptable values are ``lower``
            and ``upper``.
            For example: If set to ``lower``, prefer this::

                a { color: #fff }

            over this::

                a { color: #FFF }

        :param allow_color_shorthand:
            Whether to expand hexadecimal colors or use shorthands.
            For example: If set to ``True``, prefer this::

                b { color: #fc0 }

            over this::

                b { color: #ffcc00 }

        :param allow_leading_zero_in_dimensions:
            Add/remove leading zero in dimensions.
            For example: If set to ``False``, prefer this::

                p { padding: .5em }

            over this::

                p { padding: 0.5em }

        :param preferred_quotation:
            ``'`` or ``"`` respectively.
        :param prohibit_empty_rulesets:
            Remove all rulesets that contain nothing but spaces.
            For example: If set to ``True``, prefer this::

                a { color: red; } p { /* hey */ }

            over this::

                a { color: red; } p { /* hey */ } b { }

        :param use_space_after_colon:
            Set the number of spaces after ``:`` in declarations.
            For example: If set to ``1``, prefer this::

                a {
                    top: 0;
                    color: tomato;
                }

            over this::

                a {
                    top:0;
                    color:tomato;
                }

        :param use_space_before_colon:
            Set the number of spaces before ``:`` in declarations.
            For example: If set to ``1``, prefer this::

                a {
                    top : 0;
                    color : tomato;
                }

            over this::

                a {
                    top: 0;
                    color: tomato;
                }

        :param use_space_after_combinator:
            Set the number of spaces after the combinator.
            For example: If set to ``1``, prefer this::

                p> a { color: panda; }

            over this::

                p>a { color: panda; }

        :param use_space_before_combinator:
            Set the number of spaces before the combinator.
            For example: If set to ``1``, prefer this::

                p >a { color: panda; }

            over this::

                p>a { color: panda; }

        :param use_space_between_declarations:
            Set the number of spaces between declarations.
        :param use_space_after_opening_brace:
            Set the number of spaces after ``{``.
        :param use_space_before_closing_brace:
            Set the number of spaces before ``}``.
        :param use_space_after_selector_delimiter:
            Set the number of spaces after the selector delimiter.
            For example: If set to ``1``, prefer this::

                a, b {
                    color: panda;
                    }

            over this::

                a,b{
                    color: panda;
                    }

        :param use_space_before_selector_delimiter:
            Set the number of spaces before selector delimiter.
        :param prohibit_trailing_whitespace:
            Whether to allow trailing whitespace or not.
        :param prohibit_units_in_zero_valued_dimensions:
            Whether to remove units in zero-valued dimensions.
            For example: If set to ``True``, prefer this::

                img { border: 0 }

            over this::

                img { border: 0px }

        :param vendor_prefix_align:
            Whether to align prefixes in properties and values.
            For example: If set to ``True``, prefer this::

            a
            {
                -webkit-border-radius: 3px;
                   -moz-border-radius: 3px;
                        border-radius: 3px;
                background: -webkit-linear-gradient(top, #fff 0, #eee 100%);
                background:    -moz-linear-gradient(top, #fff 0, #eee 100%);
            }

            over this::

            a
            {
                -webkit-border-radius: 3px;
                -moz-border-radius: 3px;
                border-radius: 3px;
                background: -webkit-linear-gradient(top, #fff 0, #eee 100%);
                background: -moz-linear-gradient(top, #fff 0, #eee 100%);
            }

        :param use_lines_between_rulesets:
            Number of line breaks between rulesets or ``@rules``.
        """
        if csscomb_config:
            return None
        else:
            options = {
                'always-semicolon': enforce_semicolon if enforce_semicolon
                else None,
                'block-indent': use_block_indentation,
                'color-case': use_color_case,
                'color-shorthand': allow_color_shorthand,
                'leading-zero': allow_leading_zero_in_dimensions,
                'quotes': 'single' if preferred_quotation == "'" else 'double',
                'remove-empty-rulesets': prohibit_empty_rulesets if
                prohibit_empty_rulesets else None,
                'space-after-colon': use_space_after_colon,
                'space-before-colon': use_space_before_colon,
                'space-after-combinator': use_space_after_combinator,
                'space-before-combinator': use_space_before_combinator,
                'space-between-declarations': use_space_between_declarations,
                'space-after-opening-brace': use_space_after_opening_brace,
                'space-before-closing-brace': use_space_before_closing_brace,
                'space-after-selector-delimiter':
                    use_space_after_selector_delimiter,
                'space-before-selector-delimiter':
                    use_space_before_selector_delimiter,
                'strip-spaces': prohibit_trailing_whitespace if
                prohibit_trailing_whitespace else None,
                'unitless-zero': prohibit_units_in_zero_valued_dimensions if
                prohibit_units_in_zero_valued_dimensions else None,
                'vendor-prefix-align': vendor_prefix_align if
                vendor_prefix_align else None,
                'lines-between-rulesets': use_lines_between_rulesets
            }

            return json.dumps(options)

    @staticmethod
    def create_arguments(filename, file, config_file,
                         csscomb_config: str=''):
        """
        :param csscomb_config:
            The location of the ``.csscomb.json`` config file.
        """
        return ('--config',
                csscomb_config if csscomb_config else config_file)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='standard',
        output_format='regex',
        output_regex=r'\s*[^:]+:(?P<line>\d+):(?P<column>\d+):'
                     r'\s*(?P<message>.+)')
class JSStandardBear:
    """
    One JavaScript Style to Rule Them All.

    No decisions to make.
    No .eslintrc, .jshintrc, or .jscsrc files to manage.
    It just works.
    """

    LANGUAGES = {'JavaScript', 'JSX'}
    REQUIREMENTS = {NpmRequirement('standard', '10')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax'}
    CAN_FIX = {'Formatting'}
    SEE_MORE = 'https://standardjs.com/rules.html'

    def create_arguments(self, filename, file, config_file):
        return (filename, '--verbose')



import json

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='apertium_lint',
        output_format='regex',
        output_regex=r'(?P<severity>\w+) : (?P<message>[^:\n]+: '
                     r'(?P<origin>[^\n]+))\n[^:\n]+: (?P<line>\d+)',
        config_suffix='.json',
        severity_map={'Warning': RESULT_SEVERITY.MAJOR})
class ApertiumLintBear:
    """
    `Apertium lint` is a python module that lints out irregular yet
    acceptable constructs that may creep into files involved in the
    transfer process. The lint is designed specifically to handle
    4 file types dictionaries, transfer, modes and tagger.

    https://github.com/coala/coala-bears/issues/1697 needs to be closed
    for extending Windows support for this bear.

    Lints the file using
    `apertium_lint <https://pypi.python.org/pypi/apertium_lint>`.
    """
    LANGUAGES = {'Apertium'}
    REQUIREMENTS = {PipRequirement('apertium-lint', '0.29'),
                    PipRequirement('lxml', '3.6.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Redundancy', 'Syntax', 'Formatting', 'Unused Code',
                  'Duplication', 'Undefined Element'}
    SEE_MORE = 'https://pypi.python.org/pypi/apertium_lint'

    @staticmethod
    def create_arguments(filename, file, config_file,
                         apertiumlint_config: str=''):
        """
        :param apertiumlint_config: Path to a custom configuration file.
        """
        if apertiumlint_config:
            args = (filename, '--config', apertiumlint_config)
        else:
            args = (filename, '--config', config_file)
        return args

    @staticmethod
    def generate_config(filename, file,
                        redundant_pardef: bool=False,
                        paradigm_names: bool=True,
                        r_tag_data: bool=False,
                        repeated_attributes_pardef: bool=False,
                        repeated_entries_pardef: bool=False,
                        repeated_entries_main_section: bool=False,
                        repeated_tag_entries: bool=False,
                        monodix_transfer_direction: bool=False,
                        unused_paradigms: bool=False,
                        blank_space_detection: bool=False,
                        partially_in_lemma: bool=False,
                        unwanted_tag: bool=False,
                        bidix_transfer_direction: bool=False,
                        unwanted_white_space: bool=False,
                        compare_sdefs: bool=False,
                        verify_invariable_part: bool=False,
                        repeated_entries: bool=False,
                        repeated_entries_cat_item: bool=False,
                        repeated_entries_attr_item: bool=False,
                        unused_def_cats: bool=False,
                        check_valid_part_clip: bool=True,
                        check_valid_equal_tag: bool=True,
                        conflicting_cat_item: bool=False,
                        check_valid_position: bool=False,
                        enforce_break_tag: bool=False,
                        xsd_validation: bool=True,
                        enforce_side: bool=False,
                        macro_names: bool=False,
                        def_label_closed: bool=False,
                        validate_label_sequence: bool=False,
                        repeated_def_label: bool=True,
                        install_bool: bool=False,
                        repeated_program: bool=False,
                        validate_program: bool=False,
                        locate_file: bool=False,
                        empty_program: bool=False,
                        enforce_rules: bool=True,
                        install_switch_no: bool=True
                        ):
        """
        :param redundant_pardef:
            This issue is responsible for detecting redundant paradigms in
            monodix files.
        :param paradigm_names:
            This issue is responsible for enforcing certain rules associated
            with naming paradigms in monodix files.
        :param r_tag_data:
            This issue is responsible for detecting any inconsistency in the
            data present in the `<r>` tag in the parameter definition entries
            in monodix files.
        :param repeated_attributes_pardef:
            This issue is responsible for checking any repeated entries in
            the attributes associated with an entry in the parameter
            definition section
            n monodix files.
        :param repeated_entries_pardef:
            This issue is responsible for detecting repeated entries in the
            parameter definition entries for a given paradigm in monodix files.
        :param repeated_entries_main_section:
            This issue is responsible for detecting repeated entries in main
            section in monodix files.
        :param repeated_tag_entries:
            Same as `repeated_attributes_pardef`, but include repeated entries
            in the `<s>` and `<par>` tags in the Main Section.
        :param monodix_transfer_direction:
            This is issue is responsible for making sure that a valid transfer
            direction is supplied in monodix files.
        :param unused_paradigms:
            This issue is used for detecting unused paradigms in the parameter
            definitions section in monodix files.
        :param blank_space_detection:
            This issue detects 'extra' blank spaces that might be present in
            the entries in monodix files.
        :param partially_in_lemma:
            This issue is responsible for detecting if the parameter
            definitions that are part of lemma are written twice in monodix
            files.
        :param unwanted_tag:
            This issue detects extra/unwanted `<b/>` tags that might creep into
            entries in monodix files.
        :param bidix_transfer_direction:
            This issue is responsible for detecting and reporting incorrect
            transfer directions in bidix files.
        :param unwanted_white_space:
            This issue detects and reports white spaces that may creep into
            the entries in bidix files.
        :param compare_sdefs:
            This issue is responsible for detecting and reporting the sdefs
            that may be present in the bidix but are absent in the monodixes.
        :param verify_invariable_part:
            This issue is responsible for detecting homogeneity in definition
            of the invariable part across the monodix and the bidix.
        :param repeated_entries:
            Checks and reports repeated entries in the bidix.
        :param repeated_entries_cat_item:
            Simple check which iterates over all the `<cat-item>`s in a
            `<def­-cat>` and detects repeated `<cat-items>`
        :param repeated_entries_attr_item:
            Again another simple issue that checks for repeated `<attr-item>`s
            in `<def-attr>`.
        :param unused_def_cats:
            The lint can now detect and report redundant `<def-cat>`s that
            maybe present in the transfer rules definition but are never
            actually used.
        :param check_valid_part_clip:
            Check that the attribute listed in part='xyz' is defined
            using a `<def-attr>`. The valid parts are defined in `<def-attr>`s
            section of the transfer file.
        :param check_valid_equal_tag:
            Another common error in transfer files is trying to compare
            an attribute to a value it cannot take.
        :param conflicting_cat_item:
            This detects and reports if the same `<cat-item>` has been used in
            two or more `<def-cat>`s.
        :param check_valid_position:
            Every mention of pos='xyz' is checked to make sure that xyz
            is less than or equal to the number of elements in pattern item.
        :param enforce_break_tag:
            This function reports the if the `<b/>` tags are absent between
            two consecutive `<lu>` tags.
        :param xsd_validation:
            XSD Validation for transfer files, takes care of issues such as
            Repeated `<def-cat>`, `<def-attrs>`, and `<def-list>`.
            Valid side : `sl` or `tl` and repeated `<def-macro>`.
        :param enforce_side:
            In `<out>` the side attribute can either take the value of 'sl' or
            'tl'. With the lint in place, the user can enforce the side
            attribute to be 'tl' in all the `<out>` tags.
        :param macro_names:
            This issue detects multiple def-macro names in transfer files.
        :param def_label_closed:
            This function lints and reports if any other value has been used
            instead of the two valid ones: [true, false] in the closed
            attribute in `<def-label>` .
        :param validate_label_sequence:
            This function lints and detects and labels in <forbid> that are
            absent in `<tagset>`.
        :param repeated_def_label:
            This function detects and reports repeated def-labels in tagger
            files.
        :param install_bool:
            In modes files, the install attribute can only take one of two
            binary values, 'yes' or 'no'. This function is responsible for
            making sure that no other value is used for the same.
        :param repeated_program:
            Every modes file consits of various programs and there is always a
            chance that a certain program may unintentionally get repeated.
        :param validate_program:
            This function validates every program mentioned in the modes file.
        :param locate_file:
            This function checks and prompts incase a file defined in a
            "program" is missing in the `PWD`.
        :param empty_program:
            Not so much a risk, but this function prompts if a given "program"
            does not have any file associated with it.
        :param enforce_rules:
            This function is responsible for enforcing certain rules specific
            to given programs.
        :param install_switch_no:
            If in the definition of a certain mode, the attribute
            `install="no"`, the name should have an appropriate suffix like
            ­morph, interchunk, etc.
        """
        monodix = {'redundantPardef': {'message': 'Warning : Redundant pardef '
                                       'entry present for :',
                                       'enable': redundant_pardef},
                   'paradigmNames': {'message': 'Warning : Improper paradigm '
                                     'name found :',
                                     'enable': paradigm_names},
                   'rTagData': {'message': 'Warning : Inconsitency in <r> '
                                'text for pardef :',
                                'enable': r_tag_data},
                   'repeatedAttributesPardef': {'message': 'Warning : Repeated'
                                                ' attributes found in <s> tag '
                                                'of pardef :',
                                                'enable':
                                                repeated_attributes_pardef},
                   'repeatedEntriesPardef': {'message': 'Warning : Repeated '
                                             'entries found in pardef :',
                                             'enable':
                                             repeated_entries_pardef},
                   'repeatedEntriesMainSection': {'message': 'Warning : '
                                                  'Repeated entries found in '
                                                  'main section entry :',
                                                  'enable':
                                                  repeated_entries_main_section
                                                  },
                   'repeatedTagEntries': {'message': 'Warning : Repeated tag '
                                          'entries for <s> or <par> found in '
                                          'main section entry :',
                                          'enable': repeated_tag_entries},
                   'transferDirection': {'message': 'Warning : Invalid '
                                         'transfer direction %s found for '
                                         'entry :',
                                         'enable': monodix_transfer_direction},
                   'unusedParadigms': {'message': 'Warning : Unused paradigm'
                                       '(s) found. List of unused '
                                       'paradigm(s) :',
                                       'enable': unused_paradigms},
                   'blankSpaceDetection': {'message': 'Warning : Unused '
                                           'paradigm(s) found. List of unused '
                                           'paradigm(s) :',
                                           'enable': blank_space_detection},
                   'partiallyInLemma': {'message': 'Warning : Issue regarding '
                                        'the lemma and the paradigm associated'
                                        ' with it for the entry :',
                                        'enable': partially_in_lemma},
                   'unwantedTag': {'message': 'Warning : Unwanted <b/> tag '
                                   'found for the entry :',
                                   'enable': unwanted_tag}
                   }

        bidix = {'transferDirection': {'message': 'Warning : Invalid transfer '
                                       'direction found for entry :',
                                       'enable': bidix_transfer_direction},
                 'unwantedWhiteSpace': {'message': 'Warning : Unwanted White '
                                        'space found in the entry :',
                                        'enable': unwanted_white_space},
                 'compareSdefs': {'message': 'Warning : Certain Sdefs present '
                                  'in the bidix are absent in the monodix :',
                                  'enable': compare_sdefs},
                 'verifyInvariablePart': {'message': 'Warning : Issue '
                                          'regarding usage of <g> tag found '
                                          'for entry : ',
                                          'enable': verify_invariable_part},
                 'repeatedEntries': {'message': 'Warning : Repeated entries '
                                     'found for entry : ',
                                     'enable': repeated_entries}
                 }

        transfer = {'repeatedEntriesCatItem': {'message': 'Warning : Repeated'
                                               ' cat-item found : ',
                                               'enable':
                                               repeated_entries_cat_item},
                    'repeatedEntriesAttrItem': {'message': 'Warning : Repeated'
                                                ' attr-item found : ',
                                                'enable':
                                                repeated_entries_attr_item},
                    'unusedDefCats': {'message': 'Warning : Following def-cats'
                                      ' have not been used : ',
                                      'enable': unused_def_cats},
                    'checkValidPartClip': {'message': 'Warning : Invalid part'
                                           "='%s', found for entry : %s",
                                           'enable': check_valid_part_clip},
                    'checkValidEqualTag': {'message': "Warning : The 'part' "
                                           'does not correspond with the lit'
                                           '-tag : ',
                                           'enable': check_valid_equal_tag},
                    'conflictingCatItem': {'message': 'Warning : Repeated '
                                           'cat-item entry found for two '
                                           'def-cats : ',
                                           'enable': conflicting_cat_item},
                    'checkValidPosition': {'message': 'Warning : The position '
                                           'mentioned for a tag in the entry '
                                           'is invalid : ',
                                           'enable': check_valid_position},
                    'enforceBreakTag': {'message': 'Warning : <b/> missing '
                                        'between consecutive <lu> elements '
                                        'for element : %s, on line number'
                                        ' : %s ',
                                        'enable': enforce_break_tag},
                    'xsdValidation': {'message': 'Warning : The position '
                                      'mentioned for a tag in the entry is '
                                      'invalid : ',
                                      'enable': xsd_validation},
                    'enforceSide': {'message': 'Warning : Are you sure you '
                                    "want to specify side='sl' in the %s tag"
                                    ' on line %s',
                                    'enable': enforce_side},
                    'macroNames': {'message': 'Warning : Are you sure you '
                                   "want to ''specify multiple def-macros "
                                   "with the same n='' and npar='' values : ",
                                   'enable': macro_names}
                    }

        tagger = {'defLabelClosed': {'message': 'Warning : Invalid value for '
                                     "closed' attribute used in def-label : ",
                                     'enable': def_label_closed},
                  'validateLabelSequence': {'message': 'Warning : Invalid '
                                            'label %s used in label '
                                            'sequence. ',
                                            'enable': validate_label_sequence},
                  'repeatedDefLabel': {'message': 'Warning : Repeated def-'
                                       'label : ',
                                       'enable': repeated_def_label}
                  }

        modes = {'installBool': {'message': 'Warning : Invalid install '
                                 'attrib : %s for <mode name=\"%s\">',
                                 'enable': install_bool},
                 'repeatedProgram': {'message': 'Warning : Repeated program '
                                     'definitions in : %s',
                                     'enable': repeated_program},
                 'validateProgram': {'message': 'Warning : Invalid program '
                                     ': %s',
                                     'enable': validate_program},
                 'locateFile': {'message': 'Warning : File not found : %s',
                                'enable': locate_file},
                 'emptyProgram': {'message': 'Warning : Are you sure you '
                                  'want to keep the program %s in the '
                                  'mode %s empty?',
                                  'enable': empty_program},
                 'enforceRules': {'message': 'Warning : Certain rules not '
                                  'followed in the modes file',
                                  'enable': enforce_rules},
                 'installSwitchNo': {'message': 'Warning : Improper mode '
                                     'name %s associated with attribute'
                                     " install = 'no'",
                                     'enable': install_switch_no}
                 }

        config = {'monodix': monodix, 'bidix': bidix,
                  'transfer': transfer, 'tagger': tagger, 'modes': modes}
        return json.dumps(config, indent=2)



import json
import re

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.bearlib import deprecate_settings


@linter(executable='remark',
        use_stdout=True,
        use_stderr=True)
class MarkdownBear:
    """
    Check and correct Markdown style violations automatically.

    See <https://github.com/wooorm/remark-lint> for details about the tool
    below.
    """

    LANGUAGES = {'Markdown'}
    REQUIREMENTS = {NpmRequirement('remark-cli', '2'),
                    NpmRequirement('remark-lint', '5'),
                    NpmRequirement('remark-validate-links', '5')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}

    _output_regex = re.compile(
        r'\s*(?P<line>\d+):(?P<column>\d+)'
        r'\s*(?P<severity>warning)\s*(?P<message>.+?)(?:  .*|\n|$)')

    @staticmethod
    @deprecate_settings(bullets='markdown_bullets',
                        closed_headings='markdown_closed_headings',
                        setext_headings='markdown_setext_headings',
                        emphasis='markdown_emphasis',
                        strong='markdown_strong',
                        encode_entities='markdown_encode_entities',
                        codefence='markdown_codefence',
                        fences='markdown_fences',
                        list_indent='markdown_list_indent',
                        loose_tables='markdown_loose_tables',
                        spaced_tables='markdown_spaced_tables',
                        list_increment='markdown_list_increment',
                        horizontal_rule='markdown_horizontal_rule',
                        horizontal_rule_spaces='markdown_horizontal_'
                                               'rule_spaces',
                        horizontal_rule_repeat='markdown_horizontal_'
                                               'rule_repeat')
    def create_arguments(filename, file, config_file,
                         bullets: str='-',
                         closed_headings: bool=False,
                         setext_headings: bool=False,
                         emphasis: str='*',
                         strong: str='*',
                         encode_entities: bool=False,
                         codefence: str='`',
                         fences: bool=True,
                         list_indent: str='1',
                         loose_tables: bool=False,
                         spaced_tables: bool=True,
                         list_increment: bool=True,
                         horizontal_rule: str='*',
                         horizontal_rule_spaces: bool=False,
                         horizontal_rule_repeat: int=3,
                         max_line_length: int=None,
                         check_links: bool=False,
                         blockquote_indentation: int=2,
                         enforce_checkbox_content_indentation: bool=True,
                         code_block_style: str='consistent',
                         enforce_labels_at_eof: bool=True,
                         first_heading_level: int=None,
                         enforce_heading_level_increment: bool=False,
                         max_heading_length: int=60,
                         prohibit_duplicate_definitions: bool=True,
                         prohibit_duplicate_headings_in_section: bool=True,
                         prohibit_duplicate_headings: bool=True,
                         prohibit_empty_url: bool=True,
                         prohibit_irregular_chars_filename:
                         str='\\.a-zA-Z0-9-_',
                         prohibit_punctuations_in_heading: str='.,;:!?',
                         prohibit_html: bool=True,
                         prohibit_shortcut_reference_image: bool=True,
                         prohibit_shortcut_reference_link: bool=True,
                         use_spaces: bool=True,
                         check_undefined_references: bool=True,
                         check_unused_definition: bool=True):
        """
        :param bullets:
            Character to use for bullets in lists. Can be "-", "*" or "+".
        :param closed_headings:
            Whether to close Atx headings or not. if true, extra # marks will
            be required after the heading. eg: `## Heading ##`.
        :param setext_headings:
            Whether to use setext headings. A setext heading uses underlines
            instead of # marks.
        :param emphasis:
            Character to wrap strong emphasis by. Can be "_" or "*".
        :param strong:
            Character to wrap slight emphasis by. Can be "_" or "*".
        :param encode_entities:
            Whether to encode symbols that are not ASCII into special HTML
            characters.
        :param codefence:
            Used to find which characters to use for code fences. Can be "`" or
            "~".
        :param fences:
            Use fences for code blocks.
        :param list_indent:
            Used to find spacing after bullet in lists. Can be "1", "tab" or
            "mixed".
            - "1" - 1 space after bullet.
            - "tab" - Use tab stops to begin a sentence after the bullet.
            - "mixed" - Use "1" when the list item is only 1 line, "tab" if it
              spans multiple.
        :param loose_tables:
            Whether to use pipes for the outermost borders in a table.
        :param spaced_tables:
            Whether to add space between pipes in a table.
        :param list_increment:
            Whether an ordered lists numbers should be incremented.
        :param horizontal_rule:
            The horizontal rule character. Can be '*', '_' or '-'.
        :param horizontal_rule_spaces:
            Whether spaces should be added between horizontal rule characters.
        :param horizontal_rule_repeat:
            The number of times the horizontal rule character will be repeated.
        :param max_line_length:
            The maximum line length allowed.
        :param check_links:
            Checks if links to headings and files in markdown are valid.
        :param blockquote_indentation:
            Warn when blockquotes are either indented too much or too little.
        :param enforce_checkbox_content_indentation:
            Warn when list item checkboxes are followed by unnecessary
            whitespace.
        :param code_block_style:
            Warn when code-blocks do not adhere to a given style. Can be
            ``consistent``, ``fenced``, or ``indented``. The default value,
            ``consistent``, detects the first used code-block style, and will
            warn when a subsequent code-block uses a different style.
        :param enforce_labels_at_eof:
            Warn when definitions are not placed at the end of the file.
            For example: If set to ``True``, this will throw a warning::

                Paragraph.
                [example]: http://example.com "Example Domain"
                Another paragraph.

        :param first_heading_level:
            Warn when the first heading has a level other than a specified
            value.
            For example: If set to ``2``, this will throw a warning::

                # Bravo
                Paragraph.

        :param enforce_heading_level_increment:
            Warn when headings increment with more than 1 level at a time.
            For example: If set to ``True``, prefer this::

                # Alpha
                ## Bravo

            over this::

                # Alpha
                ### Bravo

        :param max_heading_length:
            The maximum heading length allowed. Ignores markdown syntax, only
            checks the plain text content.
        :param prohibit_duplicate_definitions:
            Warn when duplicate definitions are found.
            For example: If set to ``True``, this will throw a warning::

                [foo]: bar
                [foo]: qux

        :param prohibit_duplicate_headings_in_section:
            Warn when duplicate headings are found, but only when on the same
            level, “in” the same section.
            For example: If set to ``True``, this will throw a warning::

                ## Foxtrot
                ### Golf
                ### Golf

        :param prohibit_duplicate_headings:
            Warn when duplicate headings are found.
            For example: If set to ``True``, this will throw a warning::

                # Foo
                ## Foo
                ## [Foo](http://foo.com/bar)

        :param prohibit_empty_url:
            Warn for empty URLs in links and images.
            For example: If set to ``True``, this will throw a warning::

                [golf]().
                ![hotel]().

        :param prohibit_irregular_chars_filename:
            Warn when file names contain irregular characters: characters other
             than alpha-numericals, dashes, dots (full-stops) and underscores.
             Can take ``RegExp`` or ``string``. Any match by the given
             expression triggers a warning.
        :param prohibit_punctuations_in_heading:
            Warn when a heading ends with a group of characters. Can take a
            ``string`` that contains the group of characters.
        :param prohibit_html:
            Warn when HTML elements are used. Ignores comments, because they
            are used remark, because markdown doesn’t have native comments.
            For example: If set to ``True``, this will throw a warning::

                <h1>Hello</h1>

        :param prohibit_shortcut_reference_image:
            Warn when shortcut reference images are used.
            For example: If set to ``True``, this will throw a warning::

                ![foo]
                [foo]: http://foo.bar/baz.png

        :param prohibit_shortcut_reference_link:
            Warn when shortcut reference links are used.
            For example: If set to ``True``, this will throw a warning::

                [foo]
                [foo]: http://foo.bar/baz

        :param use_spaces:
            Warn when tabs are used instead of spaces.
        :param check_undefined_references:
            Warn when references to undefined definitions are found.
            For example: If set to ``True``, this will throw a warning::

                [bar][]

        :param check_unused_definition:
            Warn when unused definitions are found.
            For example: If set to ``True``, this will throw a warning::

                [bar]: https://example.com

        """
        remark_configs = {
            'bullet': bullets,                          # - or *
            'closeAtx': closed_headings,                # Bool
            'setext': setext_headings,                  # Bool
            'emphasis': emphasis,                       # char (_ or *)
            'strong': strong,                           # char (_ or *)
            'entities': encode_entities,                # Bool
            'fence': codefence,                         # char (~ or `)
            'fences': fences,                           # Bool
            'listItemIndent': list_indent,              # int or "tab"
                                                        # or "mixed"
            'looseTable': loose_tables,                 # Bool
            'spacedTable': spaced_tables,               # Bool
            'incrementListMarker': list_increment,      # Bool
            'rule': horizontal_rule,                    # - or * or _
            'ruleSpaces': horizontal_rule_spaces,       # Bool
            'ruleRepetition': horizontal_rule_repeat,   # int
        }
        remark_lint_configs = {
            'blockquoteIndentation': blockquote_indentation,
            'checkboxContentIndent': enforce_checkbox_content_indentation,
            'codeBlockStyle': code_block_style,
            'finalDefinition': enforce_labels_at_eof,
            'firstHeadingLevel': first_heading_level,
            'headingIncrement': enforce_heading_level_increment,
            'maximumHeadingLength': max_heading_length,
            'noDuplicateDefinitions': prohibit_duplicate_definitions,
            'noDuplicateHeadingsInSection':
                prohibit_duplicate_headings_in_section,
            'noDuplicateHeadings': prohibit_duplicate_headings,
            'noEmptyURL': prohibit_empty_url,
            'noFileNameIrregularCharacters': prohibit_irregular_chars_filename,
            'noHeadingPunctuation': prohibit_punctuations_in_heading,
            'noHTML': prohibit_html,
            'noShortcutReferenceImage': prohibit_shortcut_reference_image,
            'noShortcutReferenceLink': prohibit_shortcut_reference_link,
            'noTabs': use_spaces,
            'noUndefinedReferences': check_undefined_references,
            'noUnusedDefinitions': check_unused_definition,
        }

        if max_line_length:
            remark_lint_configs['maximumLineLength'] = max_line_length

        config_json = json.dumps(remark_configs)
        # Remove { and } as remark adds them on its own
        settings = config_json[1:-1]

        args = [filename, '--no-color', '--quiet', '--setting', settings]

        config_json = json.dumps(remark_lint_configs)
        lint = 'lint=' + config_json[1:-1]
        args += ['--use', lint]

        if check_links:
            args += ['--use', 'validate-links']

        return args

    def process_output(self, output, filename, file):
        stdout, stderr = output
        yield from self.process_output_corrected(stdout, filename, file,
                                                 RESULT_SEVERITY.NORMAL,
                                                 'The text does not comply'
                                                 ' to the set style.')
        yield from self.process_output_regex(stderr, filename, file,
                                             self._output_regex)
